# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from megatron.energon.cache.base import CachePool, FileStore, FileStoreDecoder, Lazy, MockLazy
from megatron.energon.cache.file_cache_pool import FileCacheLazy, FileStoreCachePool
from megatron.energon.cache.file_store import DecodeFileStore, SystemFileStore
from megatron.energon.cache.no_cache import DirectLazy, NoCachePool

__all__ = [
    "CachePool",
    "DecodeFileStore",
    "DirectLazy",
    "FileCacheLazy",
    "FileStore",
    "FileStoreCachePool",
    "FileStoreDecoder",
    "Lazy",
    "MockLazy",
    "NoCachePool",
    "SystemFileStore",
]



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Dict, List, Literal, Optional, Union

from megatron.energon.dataset_config import load_config
from megatron.energon.edataclass import edataclass
from megatron.energon.epathlib import EPath
from megatron.energon.errors import warn_deprecated
from megatron.energon.flavors.webdataset.metadata import check_dataset_info_present
from megatron.energon.metadataset.dataset_loader import DatasetLoader
from megatron.energon.metadataset.loader_interface import (
    DatasetBlendMode,
    DatasetLoaderInterface,
    LoadedDatasetList,
)
from megatron.energon.worker import WorkerConfig


@edataclass
class DatasetReference:
    path: Union[str, EPath]
    split_part: Optional[str] = None
    # Note: subflavor is only for legacy compatibility.
    subflavor: Optional[str] = None
    subflavors: Optional[Dict[str, Any]] = None
    shuffle_over_epochs_multiplier: Optional[int] = 1
    dataset_config: str = "dataset.yaml"
    split_config: str = "split.yaml"

    weight: float = 1.0

    _dataset: Optional[DatasetLoaderInterface] = None

    def __post_init__(self):
        if self.subflavor is not None:
            warn_deprecated(
                "subflavor is deprecated, use subflavors instead. This will be removed in a future release."
            )
            if self.subflavors is None:
                self.subflavors = {"__subflavor__": self.subflavor}
            elif "__subflavor__" not in self.subflavors:
                self.subflavors = {"__subflavor__": self.subflavor, **(self.subflavors or {})}
            self.subflavor = None

    def post_initialize(self, mds_path: Optional[EPath] = None):
        assert mds_path is not None
        if not isinstance(self.path, EPath):
            self.path = mds_path.parent / self.path
        if self.path.is_file():
            assert self.dataset_config == "dataset.yaml", "Must not set dataset_config"
            assert self.split_config == "split.yaml", "Must not set split_config"
            self._dataset = load_config(
                self.path,
                default_type=Metadataset,
                default_kwargs=dict(path=self.path),
            )
            self._dataset.post_initialize()
        elif check_dataset_info_present(self.path):
            self._dataset = DatasetLoader(path=self.path)
            self._dataset.post_initialize()
        else:
            raise FileNotFoundError(self.path)

    def get_datasets(
        self,
        *,
        training: bool,
        split_part: Union[Literal["train", "val", "test"], str],
        worker_config: WorkerConfig,
        subflavors: Optional[Dict[str, Any]] = None,
        shuffle_over_epochs_multiplier: Optional[int] = 1,
        **kwargs,
    ) -> LoadedDatasetList:
        if self.subflavors is not None:
            subflavors = {**self.subflavors, **(subflavors or {})}
        assert self._dataset is not None

        if shuffle_over_epochs_multiplier is None or self.shuffle_over_epochs_multiplier is None:
            # If no shuffling is requested, this has override priority.
            new_shuffle_over_epochs_multiplier = None
        elif shuffle_over_epochs_multiplier == -1 or self.shuffle_over_epochs_multiplier == -1:
            # Next priority is sampling without replacement.
            new_shuffle_over_epochs_multiplier = -1
        else:
            # Otherwise, multiply the shuffle over epochs multiplier.
            new_shuffle_over_epochs_multiplier = (
                shuffle_over_epochs_multiplier * self.shuffle_over_epochs_multiplier
            )

        return self._dataset.get_datasets(
            training=training,
            split_part=self.split_part or split_part,
            worker_config=worker_config,
            subflavors=subflavors,
            shuffle_over_epochs_multiplier=new_shuffle_over_epochs_multiplier,
            **kwargs,
        )


@edataclass
class MetadatasetBlender:
    """Internal blending of the dataset."""

    datasets: List[DatasetReference]

    def post_initialize(self, mds_path: Optional[EPath] = None):
        assert mds_path is not None
        for dataset in self.datasets:
            dataset.post_initialize(mds_path)

    def get_datasets(
        self,
        *,
        training: bool,
        split_part: Union[Literal["train", "val", "test"], str],
        worker_config: WorkerConfig,
        subflavors: Optional[Dict[str, Any]] = None,
        shuffle_over_epochs_multiplier: Optional[int] = 1,
        **kwargs,
    ) -> LoadedDatasetList:
        sum_weight = sum(dataset.weight for dataset in self.datasets)
        datasets = []
        for dataset in self.datasets:
            inner_result = dataset.get_datasets(
                training=training,
                split_part=split_part,
                worker_config=worker_config,
                subflavors=subflavors,
                shuffle_over_epochs_multiplier=shuffle_over_epochs_multiplier,
                **kwargs,
            )
            if inner_result.blend_mode not in (
                DatasetBlendMode.NONE,
                DatasetBlendMode.DATASET_WEIGHT,
            ):
                raise ValueError(
                    "Can only blend datasets which are of the same blend mode. Cannot mix blend with blend_epochized."
                )
            for loaded_dataset in inner_result.datasets:
                if inner_result.blend_mode == DatasetBlendMode.DATASET_WEIGHT:
                    assert isinstance(loaded_dataset.weight, float)
                else:
                    assert loaded_dataset.weight is None
                    loaded_dataset.weight = 1.0
                loaded_dataset.weight = loaded_dataset.weight * dataset.weight / sum_weight
                datasets.append(loaded_dataset)
        return LoadedDatasetList(
            blend_mode=DatasetBlendMode.DATASET_WEIGHT,
            datasets=datasets,
        )


class Metadataset(DatasetLoaderInterface):
    """Main entry for metadataset."""

    _path: EPath
    _splits: Dict[str, MetadatasetBlender]

    def __init__(
        self,
        path: Union[EPath, str],
        splits: Dict[str, MetadatasetBlender],
    ):
        """Create the metadataset"""
        self._path = EPath(path)
        self._splits = splits

    def post_initialize(self, mds_path: Optional[EPath] = None):
        assert mds_path is None
        for split in self._splits.values():
            split.post_initialize(self._path)

    def get_datasets(
        self,
        *,
        training: bool,
        split_part: Union[Literal["train", "val", "test"], str],
        worker_config: WorkerConfig,
        subflavors: Optional[Dict[str, Any]] = None,
        shuffle_over_epochs_multiplier: Optional[int] = 1,
        **kwargs,
    ) -> LoadedDatasetList:
        return self._splits[split_part].get_datasets(
            training=training,
            split_part=split_part,
            worker_config=worker_config,
            subflavors=subflavors,
            shuffle_over_epochs_multiplier=shuffle_over_epochs_multiplier,
            **kwargs,
        )



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Dict, Iterator, List, Tuple, TypeVar

import torch

from megatron.energon.flavors.base_dataset import SavableDataset, add_sample_restore_key
from megatron.energon.rng import WorkerRng
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers.base import BaseWrapperDataset

T_sample = TypeVar("T_sample")


class BlendDataset(BaseWrapperDataset[T_sample, T_sample]):
    """
    This dataset wrapper blends multiple iterable datasets together give a weighting.
    The datasets may be infinite. This dataset is always infinite.
    """

    weights: Tuple[float, ...]
    exhausted: List[bool]
    _worker_rng: WorkerRng

    _savable_fields = ("exhausted", "_worker_rng")

    def __init__(
        self,
        *dataset_weights: Tuple[SavableDataset[T_sample], float],
        worker_config: WorkerConfig,
    ):
        """Construct a BlendDataset.

        Args:
            dataset_weights: Each argument should be a tuple of (dataset, weight) with a weight
                between 0 and 1. The output samples are sampled from the input datasets with the
                given probabilities.
            worker_config: Configuration for the workers.
        """
        # datasets = [dataset for dataset, _weight in dataset_weights]
        self.datasets, self.weights = zip(*dataset_weights)

        super().__init__(self.datasets, worker_config=worker_config)

        self.dataset_weights = dataset_weights
        self.reset_state_own()

    def reset_state_own(self) -> None:
        self._worker_rng = WorkerRng(self.worker_config)
        self.exhausted = [False] * len(self.weights)

    def __len__(self) -> int:
        # Give the number of samples in inner datasets, disregarding the weight
        return sum(len(dataset) for dataset, weight in self.dataset_weights)

    def __iter__(self) -> Iterator[T_sample]:
        assert self.worker_has_samples(), "Cannot blend all empty datasets"

        # Create a list of datasets and their weights, but
        # set the weight to 0 if the dataset has no samples on this worker.

        dataset_iters = []
        weights = []
        for idx, (dataset, weight) in enumerate(self.dataset_weights):
            assert weight > 0, "All blending weights must be > 0"

            if dataset.worker_has_samples():
                dataset_iters.append(iter(dataset))
                weights.append(weight)
            else:
                dataset_iters.append(None)
                weights.append(0)

        weights = torch.tensor(weights, dtype=torch.float32)
        if weights.sum() == 0:
            raise RuntimeError(
                "There is a worker with no samples in any of the blended datasets. "
                "This can happen if you have a lot of workers and your dataset is too small. "
                "Currently this case is not supported."
            )

        # Some may already be exhausted on this worker when restoring a state.
        for idx, exhausted in enumerate(self.exhausted):
            if exhausted:
                weights[idx] = 0
                dataset_iters[idx] = None

        while True:
            ds_idx = self._worker_rng.choice_idx(probs=weights)

            if dataset_iters[ds_idx] is None:
                if all(dataset_iter is None for dataset_iter in dataset_iters):
                    break
                continue
            try:
                sample = next(dataset_iters[ds_idx])
            except StopIteration:
                dataset_iters[ds_idx] = None
                weights[ds_idx] = 0
                self.exhausted[ds_idx] = True
                if all(dataset_iter is None for dataset_iter in dataset_iters):
                    break
            else:
                yield add_sample_restore_key(sample, ds_idx, src=self)

        self.exhausted = [False] * len(self.dataset_weights)

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset_weights": [
                (dataset.config(), weight) for dataset, weight in self.dataset_weights
            ],
            "worker_config": self.worker_config.config(),
        }

    def __str__(self):
        return f"BlendDataset(dataset_weights={self.dataset_weights})"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import multiprocessing
from abc import ABC, abstractmethod
from typing import Any, Callable, Generic, Iterable, List, Optional, TypeVar

T_result = TypeVar("T_result")
T_aggregation_data = TypeVar("T_aggregation_data")
T_input_data = TypeVar("T_input_data")


class BaseAggregator(ABC, Generic[T_aggregation_data, T_result]):
    """
    Base class for a user-defined aggregator.
    Implement on_start, on_item, and on_finish to handle aggregator logic.
    """

    def on_start(self, aggregator_pool: AggregatorPool) -> None:
        """
        Called exactly once in the aggregator process before receiving any items.
        """
        pass

    @abstractmethod
    def on_item(self, item: T_aggregation_data, aggregator_pool: AggregatorPool) -> None:
        """
        Called for each item produced by the workers.
        """
        ...

    def on_finish(self, aggregator_pool: AggregatorPool) -> None:
        """
        Called once when all workers have signaled completion (i.e. all items are processed).
        """
        pass

    def get_final_result_data(self) -> T_result:
        """
        Called after on_finish to retrieve any final data produced by the aggregator.
        """
        return None


class AggregatorPool(Generic[T_input_data, T_aggregation_data, T_result]):
    """
    A pool that manages multiple worker processes sending results to
    a single aggregator process.

    The user must provide:
      - user_produce_data(task) -> yields items (streaming results)
      - aggregator: an instance of a class derived from BaseAggregator
                    which implements on_start, on_item, on_finish, etc.
    """

    num_workers: int
    user_produce_data: Callable[[T_input_data], Iterable[Any]]
    aggregator: BaseAggregator[T_aggregation_data, T_result]

    task_queue: multiprocessing.Queue[Optional[T_input_data]]
    result_queue: multiprocessing.Queue[Optional[T_aggregation_data]]

    def __init__(
        self,
        num_workers: int,
        user_produce_data: Callable[[T_input_data], Iterable[Any]],
        aggregator: BaseAggregator[T_aggregation_data, T_result],
    ) -> None:
        """
        Args:
            num_workers: Number of worker processes.
            user_produce_data: Function that takes a task and yields items (the "large" data stream).
            aggregator: An instance of a user-defined class for handling aggregator logic.
        """
        self.num_workers = num_workers
        self.user_produce_data = user_produce_data
        self.aggregator = aggregator

        # Queues for tasks and results
        self.task_queue = multiprocessing.Queue()
        self.result_queue = multiprocessing.Queue()

        # Queue to pass final aggregator data back to the main process
        self._final_result_data_queue = multiprocessing.Queue()

        # Will store whatever is pulled from _final_data_queue in close()
        self._aggregator_final_result_data: Optional[Any] = None

    def _worker(self, worker_id: int) -> None:
        """Function that runs inside each worker process."""
        while True:
            task = self.task_queue.get()
            if task is None:
                # No more tasks, signal aggregator that this worker is done
                break

            # Produce data in a streaming fashion
            for item in self.user_produce_data(task):
                self.result_queue.put(item)

        # After finishing all tasks, send a sentinel to the aggregator
        self.result_queue.put(None)

    def _aggregator_run(self) -> T_result:
        """
        Function that runs in the aggregator process.
        Keeps reading items from result_queue.
        - If an item is None, that means a worker finished all of its tasks.
        - Otherwise, call aggregator.on_item(...) with that item.
        """
        # Let the aggregator do any initialization it needs
        self.aggregator.on_start(self)

        finished_workers = 0

        while finished_workers < self.num_workers:
            item = self.result_queue.get()
            if item is None:
                # A worker has finished all of its tasks
                finished_workers += 1
            else:
                # Process the item in the aggregator
                self.aggregator.on_item(item, self)

        # All workers done, aggregator can finalize
        self.aggregator.on_finish(self)

        # After finishing, serialize the aggregator's final data
        return self.aggregator.get_final_result_data()

    def submit_task(self, task: T_input_data) -> None:
        """
        Submit a task to be processed by a worker.
        """
        self.task_queue.put(task)

    def process(self) -> T_result:
        """
        Starts the worker processes and runs the aggregation in the main process.
        Waits for all workers to finish and retrieves the aggregator's final data.
        """
        workers: List[multiprocessing.Process] = []

        # Start worker processes
        for w_id in range(self.num_workers):
            p = multiprocessing.Process(target=self._worker, args=(w_id,), daemon=True)
            p.start()
            workers.append(p)

        # Send the sentinel (None) to each worker
        for _ in range(self.num_workers):
            self.task_queue.put(None)

        # Now wait for aggregator to finish reading the queue
        result = self._aggregator_run()

        # Wait for all workers to finish
        for p in workers:
            p.join()

        return result



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import functools
import json
import logging
import random
import re
import tarfile
import uuid
from dataclasses import asdict
from pathlib import Path
from typing import (
    Any,
    Callable,
    Container,
    Dict,
    Generator,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

import braceexpand
import yaml

from megatron.energon import __version__
from megatron.energon.bracecollapse import collapse
from megatron.energon.edataclass import edataclass
from megatron.energon.epathlib import EPath
from megatron.energon.eyaml import load_yaml
from megatron.energon.flavors.webdataset.aggregator_pool import AggregatorPool, BaseAggregator
from megatron.energon.flavors.webdataset.config import MAIN_FOLDER_NAME, skip_meta_re, split_name_re
from megatron.energon.flavors.webdataset.indexing import SqliteIndexWriter
from megatron.energon.flavors.webdataset.itar import TarIndexWriter
from megatron.energon.flavors.webdataset.structs import ShardInfo, WebdatasetInfo, WebdatasetSplits
from megatron.energon.typed_converter import to_json_object

logger = logging.getLogger(__name__)

T = TypeVar("T", covariant=True)


@edataclass
class IndexAggregatable:
    """
    A base class for all objects that can be returned/yielded by `_preprocess_tar` and
    received by `SqliteIndexWriterAggregator.on_item`.
    """

    ...


@edataclass
class IndexSample(IndexAggregatable):
    tar_file_id: int
    sample_key: str
    sample_index: int
    byte_offset: int
    byte_size: int


@edataclass
class IndexSamplePart(IndexAggregatable):
    tar_file_id: int
    sample_index: int
    part_name: str
    content_byte_offset: int
    content_byte_size: int


@edataclass
class IndexShardInfo(IndexAggregatable):
    shard_info: ShardInfo
    parts: Set[str]


class SqliteIndexWriterAggregator(
    BaseAggregator[
        Tuple[ShardInfo, Set[str]], Tuple[List[ShardInfo], Set[str], bool, List[Tuple[str, int]]]
    ]
):
    sqlite_path: EPath
    total_tasks: int
    progress_fn: Optional[Callable]
    writer: Optional[SqliteIndexWriter]
    had_update: bool
    shards: List[ShardInfo]
    found_parts: Set[str]
    prog_iter: Iterator

    def __init__(
        self,
        sqlite_path: EPath,
        total_tasks: int,
        progress_fn: Optional[Callable[[Iterator[Any], int], Iterator[T]]] = None,
    ):
        self.sqlite_path = sqlite_path
        self.total_tasks = total_tasks
        self.writer = None
        self.had_update = False
        self.shards = []
        self.found_parts = set()

        if progress_fn is not None:
            self.prog_iter = progress_fn(iter(range(self.total_tasks)), self.total_tasks)
        else:
            self.prog_iter = iter(range(self.total_tasks))

    def on_start(self, aggregator_pool: AggregatorPool) -> None:
        self.writer = SqliteIndexWriter(self.sqlite_path)

    def on_item(
        self,
        item: IndexAggregatable,
        aggregator_pool: AggregatorPool,
    ) -> None:
        assert self.writer is not None, "Writer is not initialized."
        if isinstance(item, IndexSample):
            self.writer.append_sample(**asdict(item))
            self.had_update = True
        elif isinstance(item, IndexSamplePart):
            self.writer.append_part(**asdict(item))
        elif isinstance(item, IndexShardInfo):
            # This is a (shard_info, parts) tuple
            next(self.prog_iter)

            shard_info, cur_parts = item.shard_info, item.parts
            assert shard_info.count != 0, f"Shard {shard_info.name} has no samples."
            self.shards.append(shard_info)
            if len(self.found_parts) < 50:
                self.found_parts.update(cur_parts)

    def on_finish(self, aggregator_pool: AggregatorPool) -> None:
        assert self.writer is not None, "Writer is not initialized."
        self.writer.close()

    def get_final_result_data(
        self,
    ) -> Tuple[List[ShardInfo], Set[str], bool, List[Tuple[str, int]]]:
        assert self.writer is not None, "Writer is not initialized."
        return self.shards, self.found_parts, self.had_update, self.writer.duplicates


class WebdatasetPreparator:
    @staticmethod
    def _preprocess_tar(
        path: str,
        shard_to_idx: Dict[str, int],
        parent_path: EPath,
        max_parts: int,
    ) -> Generator[IndexAggregatable, None, None]:
        """Process a single tar file, i.e. read the tarinfos, generate the tar index and return
        stats.
        This method is passed to the `user_produce_data` argument of AggregatorPool.

        Args:
            path: Path to the tar file.
            shard_to_idx: Mapping from shard path to its index
            parent_path: Root path of the dataset.
            max_parts: Maximum number of different parts to return

        Returns:
            A generator of items that will be processed by SqliteIndexWriterAggregator.
            See method `on_item` of SqliteIndexWriterAggregator.
            The items are either:
            - A sample dictionary with information about the offset, key etc.
            - Or a tuple of shard info and a set of found parts for statistics.
        """
        shard_info = ShardInfo(name=path, path=parent_path / path, count=0)

        try:
            # Note: Write to .tmp file first, then remove .tmp extension, to make sure only complete
            # files are used.
            tar: tarfile.TarFile
            with shard_info.path.open("rb") as f:
                with (
                    tarfile.open(fileobj=f, mode="r:*") as tar,
                    TarIndexWriter(shard_info.path) as iw,
                ):
                    count = 0

                    # The parts set is used to collect various file endings that are
                    # available in the dataset. This is used for the interactive prepare wizard.
                    parts = set()

                    last_base_name = None
                    member: tarfile.TarInfo

                    next_index_sample = None

                    for member in tar:
                        if not member.isreg():
                            continue
                        if member.name is None:
                            continue
                        if skip_meta_re.match(member.name):
                            continue

                        name_match = split_name_re.match(member.name)
                        if name_match is None:
                            continue

                        base_name = name_match.group(1)
                        if len(parts) < max_parts:
                            parts.add(name_match.group(2))

                        if last_base_name != base_name:
                            iw.append(member.offset)

                            if next_index_sample is not None:
                                next_index_sample["byte_size"] = (
                                    member.offset - next_index_sample["byte_offset"]
                                )
                                yield IndexSample(**next_index_sample)

                            next_index_sample = dict(
                                tar_file_id=shard_to_idx[path],
                                sample_key=base_name,
                                sample_index=count,
                                byte_offset=member.offset,
                            )
                            last_base_name = base_name
                            count += 1

                        # Yield this part of the sample to the aggregator
                        yield IndexSamplePart(
                            tar_file_id=shard_to_idx[path],
                            sample_index=count - 1,
                            part_name=name_match.group(2),
                            content_byte_offset=member.offset_data,
                            content_byte_size=member.size,
                        )

                    shard_info.count = count
                    iw.append(tar.offset)
                    if next_index_sample is not None:
                        next_index_sample["byte_size"] = (
                            tar.offset - next_index_sample["byte_offset"]
                        )
                        yield IndexSample(**next_index_sample)
            yield IndexShardInfo(shard_info=shard_info, parts=parts)
            return
        except BaseException:
            logger.exception(f"Shard failed to load: {path!r}. Skipping it.")
            yield IndexShardInfo(shard_info=shard_info, parts=set())
            return

    @staticmethod
    def iter_dataset_content(
        path: Union[str, EPath],
        extract_keys: Container[str] = (),
    ) -> Generator[Dict[str, Any], None, None]:
        """
        Yield example dataset content for a few samples.

        Args:
            path: Path to the tar file.
        """
        path = EPath(path)
        with path.open("rb") as f:
            tar: tarfile.TarFile
            with tarfile.open(fileobj=f, mode="r:*") as tar:
                last_base_name = None
                sample = {}
                member: tarfile.TarInfo
                for member in tar:
                    if not member.isreg():
                        continue
                    if member.name is None:
                        continue
                    if skip_meta_re.match(member.name):
                        continue

                    name_match = split_name_re.match(member.name)
                    if name_match is None:
                        continue

                    base_name = name_match.group(1)
                    if last_base_name != base_name:
                        if sample:
                            yield sample
                        sample = {}
                        last_base_name = base_name
                    if name_match:
                        if name_match.group(2) in extract_keys:
                            sample[name_match.group(2)] = tar.extractfile(member).read()
                        else:
                            sample[name_match.group(2)] = None
                if sample:
                    yield sample

    @classmethod
    def prepare_dataset(
        cls,
        parent_path: Union[Path, EPath],
        paths: List[str],
        *,
        split_parts_ratio: Optional[List[Tuple[str, float]]] = None,
        split_parts_patterns: Optional[List[Tuple[str, str]]] = None,
        split_config: str = "split.yaml",
        shuffle_seed: Optional[int] = 42,
        progress_fn: Callable[[Iterator[Any], int], Iterator[T]] = (lambda x, y: x),
        workers: int = 32,
        tar_index_only: bool = False,
    ) -> Tuple[Set[str], List[Tuple[str, int]]]:
        """
        Preprocess the shards and write the split config. Preprocessing is done in parallel.
        Counts the number of samples in each shard.

        Args:
            parent_path: Common parent path for the shards
            paths: Paths to the shards
            split_parts_ratio: Names of splits and their ratio (will be normalized)
            split_parts_patterns: Names of splits and their path patterns
            split_config: Filename for the split config (`parent_path / '.nv-meta' / split_config`), may be yaml or json
            shuffle_seed: Seed for shuffling shards before splitting into split_parts. None to
                disable.
            progress_fn: Callback for progress bar
            workers: Number of parallel workers for reading each shard
            tar_index_only: Only create tar-index, then exit

        Returns:
            The set of all parts found in the shards. But at most 50.
        """
        parent_path = EPath(parent_path)

        paths = [path for path in paths for path in braceexpand.braceexpand(path)]

        # Construct a mapping from relative shard path to its index
        shard_to_idx = {path: idx for idx, path in enumerate(paths)}

        (parent_path / MAIN_FOLDER_NAME).mkdir(exist_ok=True)

        aggregator = SqliteIndexWriterAggregator(
            parent_path / MAIN_FOLDER_NAME / "index.sqlite",
            total_tasks=len(paths),
            progress_fn=progress_fn,
        )

        process_tar = functools.partial(
            cls._preprocess_tar,
            shard_to_idx=shard_to_idx,
            parent_path=parent_path,
            max_parts=50,
        )

        pool = AggregatorPool(
            num_workers=workers,
            user_produce_data=process_tar,
            aggregator=aggregator,
        )

        for path in paths:
            pool.submit_task(path)

        shards, found_parts, had_update, duplicates = pool.process()

        if had_update:
            logger.info("Regenerating dataset UUID...")
            with (parent_path / MAIN_FOLDER_NAME / "index.uuid").open("w") as f:
                f.write(str(uuid.uuid4()))

        json_info_config = parent_path / MAIN_FOLDER_NAME / ".info.json"
        yaml_info_config = parent_path / MAIN_FOLDER_NAME / ".info.yaml"

        if tar_index_only:
            if yaml_info_config.is_file() and not json_info_config.is_file():
                # Convert legacy .info.yaml to .info.json
                with json_info_config.open("w") as f:
                    json.dump(load_yaml(yaml_info_config.read_bytes()), f, indent=2)

            return found_parts, duplicates

        assert len(shards) == len(shard_to_idx), (
            f"Lengths of shards and shard_to_idx do not match: {len(shards)} != {len(shard_to_idx)}"
        )

        # Sort the shards according to the order in the input list
        shards.sort(key=lambda shard: shard_to_idx[shard.name])

        # Save info
        assert [shard.name for shard in shards] == list(shard_to_idx.keys()), (
            "Shards are not in the same order as in the input list."
        )

        info = WebdatasetInfo(
            energon_version=__version__,
            shard_counts={shard.name: shard.count for shard in shards},
        )
        print(f"Saving info to {json_info_config}")

        with json_info_config.open("w") as wf:
            json.dump(to_json_object(info), wf, indent=2)

        if yaml_info_config.is_file():
            # If a .info.yaml existed previously, let's also update it
            # to keep them in sync
            with yaml_info_config.open("w") as wf:
                yaml.dump(to_json_object(info), wf)

        if split_parts_ratio is not None:
            # Normalize ratio
            total_ratio = sum(split_ratio for _, split_ratio in split_parts_ratio)
            split_parts_ratio = [
                (split_part, split_ratio / total_ratio)
                for split_part, split_ratio in split_parts_ratio
            ]
            # Sample from shards based on the split ratio from split parts
            split_shards = {}
            if shuffle_seed is not None:
                random.Random(shuffle_seed).shuffle(shards)
            split_total = 0
            split_offset = 0
            for split_part, split_ratio in split_parts_ratio:
                split_total += split_ratio
                split_end = int(len(shards) * split_total)
                split_shards[split_part] = [shard.name for shard in shards[split_offset:split_end]]
                split_offset = split_end
        else:
            assert split_parts_patterns is not None, (
                "Require either split_parts_ratio or split_parts_patterns"
            )
            # Sample from shards based on the split patterns from split parts
            split_shards = {}
            for split_part, split_pattern in split_parts_patterns:
                patterns = [
                    re.compile(pattern) for pattern in braceexpand.braceexpand(split_pattern)
                ]
                split_shards[split_part] = [
                    shard.name
                    for shard in shards
                    if any(pattern.match(shard.name) for pattern in patterns)
                ]

        # Optimize the split parts by trying to bracecollapse the shard names
        print("Collapsing split parts... ", flush=True, end="")
        for split_part in split_shards:
            split_shards[split_part] = collapse(split_shards[split_part], keep_order=True)
        print("Done", flush=True)

        # Save split config
        splits_config = WebdatasetSplits(split_parts=split_shards)
        with (parent_path / MAIN_FOLDER_NAME / split_config).open("w") as wf:
            if split_config.endswith(".yaml"):
                yaml.dump(to_json_object(splits_config), wf, sort_keys=False)
            elif split_config.endswith(".json"):
                json.dump(to_json_object(splits_config), wf, indent=2)
            else:
                raise ValueError(f"Invalid split config extension: {split_config}")

        return found_parts, duplicates



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import List, Optional

import torch

from megatron.energon.edataclass import edataclass
from megatron.energon.epathlib.epath import EPath
from megatron.energon.errors import warn_deprecated
from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.webdataset import DefaultDecoderWebdatasetFactory


@edataclass
class SimilarityInterleavedSample(Sample):
    """Sample type for interleaved media such as text with images, but without image-text alignment.
    That alignment has to be assigned from the similarity matrix."""

    #: The images of the sequence
    images: List[torch.Tensor]
    #: The texts of the sequence
    texts: List[str]
    #: The optional audio samples of the sequence
    audio: Optional[List[torch.Tensor]] = None
    #: The optional video frames of the sequence
    video: Optional[List[torch.Tensor]] = None

    #: Similarity matrix between image and text entries in the sequence
    similarity_matrix: Optional[torch.Tensor] = None

    #: The index within texts representing the sentence that this image is matched to
    matched_text_indices: Optional[List[int]] = None


class SimilarityInterleavedWebdataset(DefaultDecoderWebdatasetFactory[SimilarityInterleavedSample]):
    __sample_type__ = SimilarityInterleavedSample

    def __init__(self, path: EPath, **kwargs):
        warn_deprecated(
            f"{type(self)} is deprecated, use the default instead and set the sample_type:\n"
            f"To convert, update your {path}/.nv-meta/dataset.yaml to:\n"
            f"# remove top-level __module__ and __class__\n"
            f"sample_type:\n"
            f"  __module__: megatron.energon\n"
            f"  __class__: {self.__sample_type__.__name__}\n"
            f"# Keep the remaining content"
        )
        super().__init__(path, **kwargs)



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Callable, Literal

import webdataset

from megatron.energon.av import AVWebdatasetDecoder
from megatron.energon.cache import FileStoreDecoder
from megatron.energon.flavors.webdataset.config import split_name_re

try:
    import filetype
except ImportError:
    filetype = None

ImageDecoderType = Literal[
    "l8",
    "rgb8",
    "rgba8",
    "l",
    "rgb",
    "rgba",
    "torchl8",
    "torchrgb8",
    "torchrgba8",
    "torchl",
    "torchrgb",
    "torch",
    "torchrgba",
    "pill",
    "pil",
    "pilrgb",
    "pilrgba",
]

AVDecoderType = Literal["torch", "AVDecoder", "pyav"]


class GuessingHandlerWrapper:
    """A wrapper that guesses the extension of the file using the `filetype` package."""

    def __init__(self, handler: Callable[[str, bytes], Any]):
        """
        Wraps a handler to guess the extension of the file using the `filetype` package.

        Args:
            handler: The handler to wrap.
        """
        self.handler = handler
        if filetype is None:
            raise ImportError("filetype is not installed. Install it with `pip install filetype`.")

    def __call__(self, key: str, data: bytes) -> Any:
        """The handler that guesses the extension of the file using the `filetype` package, then calls the delegate handler."""
        kind = filetype.guess(data)
        if kind is not None:
            key = kind.extension
        return self.handler(key, data)

    @staticmethod
    def wrap(
        active: bool, handlers: list[Callable[[str, bytes], Any]]
    ) -> list[Callable[[str, bytes], Any]]:
        """
        Wraps a list of handlers to guess the extension of the file using the `filetype` package.

        Args:
            active: Whether to wrap the handlers.
            handlers: The handlers to wrap.

        Returns:
            The list of wrapped handlers.
        """
        if not active:
            return handlers
        return [GuessingHandlerWrapper(handler) for handler in handlers]


class SampleDecoder(FileStoreDecoder):
    """The default decoder for webdataset samples."""

    def __init__(
        self,
        *,
        image_decode: ImageDecoderType = "torchrgb",
        av_decode: AVDecoderType = "AVDecoder",
        video_decode_audio: bool = False,
        guess_content: bool = False,
    ):
        """
        Args:
            image_decode: This defines the decoding results.
            av_decode: If "AVDecoder", returns an AVDecoder instance for flexible decoding. If "torch",
                returns decoded VideoData.
            video_decode_audio: Whether to decode audio from video files.
            guess_content: Whether to guess the contents of the file using the `filetype` package.
        """
        self._config = dict(
            image_decode=image_decode,
            av_decode=av_decode,
            video_decode_audio=video_decode_audio,
            guess_content=guess_content,
        )
        self._decoder = webdataset.autodecode.Decoder(
            GuessingHandlerWrapper.wrap(
                guess_content,
                [
                    webdataset.autodecode.imagehandler(image_decode),
                    AVWebdatasetDecoder(
                        video_decode_audio=video_decode_audio,
                        av_decode=av_decode,
                    ),
                ],
            ),
        )

    def decode(self, fname: str, raw: bytes) -> Any:
        m = split_name_re.match(fname)
        if m:
            cur_base_name, ext = m.groups()
        else:
            # If the file name does not match the expected format, use the file name as is.
            # The handlers may still handle the file correctly (e.g. if guessing by content).
            cur_base_name = fname
            ext = "."

        return self(
            {
                "__key__": cur_base_name,
                ext: raw,
            }
        )[ext]

    def __call__(self, sample: dict) -> dict:
        return self._decoder(sample)

    def config(self) -> dict:
        return self._config



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Sequence, Union

from megatron.energon.cache import FileStore
from megatron.energon.edataclass import edataclass
from megatron.energon.epathlib import EPath
from megatron.energon.flavors.base_dataset import BaseCoreDatasetFactory
from megatron.energon.worker import WorkerConfig


class DatasetBlendMode(Enum):
    """Determines how the the datasets are to be blended. Either by using the associated number as
    the weight for sampling from that dataset, or alternatively by using the number as the number
    of repetitions for samples in that dataset in one epoch (effectively, that corresponds to the
    weight for samples)."""

    NONE = "none"
    DATASET_WEIGHT = "dataset_weight"
    SAMPLE_REPETITIONS = "sample_repetitions"


@edataclass
class LoadedDataset:
    dataset: BaseCoreDatasetFactory
    weight: Union[float, int, None] = None
    repetitions: Union[float, int, None] = None
    aux: Optional[Dict[str, FileStore]] = None


@edataclass
class LoadedDatasetList:
    datasets: List[LoadedDataset]
    blend_mode: DatasetBlendMode = DatasetBlendMode.NONE


class DatasetLoaderInterface(ABC):
    """General interface for a dataset loader."""

    @abstractmethod
    def post_initialize(self, mds_path: Optional[EPath] = None):
        """Called to finally initialize the dataset."""
        ...

    @abstractmethod
    def get_datasets(
        self,
        *,
        training: bool,
        split_part: Union[Literal["train", "val", "test"], str],
        worker_config: WorkerConfig,
        subflavors: Optional[Dict[str, Any]] = None,
        shuffle_over_epochs_multiplier: Optional[int] = 1,
        **kwargs,
    ) -> LoadedDatasetList:
        """
        Calls :func:`megatron.energon.dataset_config.get_dataset_from_config` (loads the raw dataset)
        for all innermost datasets and resolves their relative weights to absolute weights.

        Args:
            training: If true, apply training randomization.
            split_part: Default split part to use.
            worker_config: Worker configuration to use.
            subflavors: Set the default subflavors for all datasets.
            shuffle_over_epochs_multiplier: Multiply the inner datasets
                `shuffle_over_epochs(_multiplier)` by this factor. E.g. if the inner dataset
                has `shuffle_over_epochs_multiplier=2` and this function has
                `shuffle_over_epochs_multiplier=3`, the inner dataset will be shuffled
                over 6 epochs. Shuffling over `n` epochs guarantees that each sample is seen
                exactly `n` times in `n` epochs of the inner dataset. Use -1 for shuffling over
                an infinite number of epochs (effectively, this will draw shard slices with
                replacement).
            **kwargs: Additional arguments to the dataset constructor.

        Returns:
            The dataset blending mode and the instantiated core datasets with their weights/repetitions.
        """
        ...

    def prepare(self, split_part: Optional[str] = None) -> Sequence[EPath]:
        """
        Prepares the loader by creating caches and other necessary structures on disk.

        Args:
            split_part: Name of the split to load.

        Returns:
            List of paths to the cache paths. This is used for cleanup.
        """
        return ()



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional

import torch

from megatron.energon.edataclass import edataclass


@edataclass
class AVData:
    #: A list of video clips in the shape (frames, channels, h, w)
    video_clips: Optional[list[torch.Tensor]]
    #: The timestamps for the video clips. List of tuples (start, end) in seconds
    video_timestamps: Optional[list[tuple[float, float]]]
    #: A list of audio clips in the shape (channels, samples)
    audio_clips: Optional[list[torch.Tensor]]
    #: The timestamps for the audio clips. List of tuples (start, end) in seconds
    audio_timestamps: Optional[list[tuple[float, float]]]



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause


import logging
import traceback
from typing import Optional

import click
import tqdm

from megatron.energon import WorkerConfig, get_loader, get_val_dataset
from megatron.energon.epathlib import EPath
from megatron.energon.flavors.webdataset.empty_dataset_error import EmptyDatasetError
from megatron.energon.flavors.webdataset.itar import TarIndexReader
from megatron.energon.flavors.webdataset.metadata import get_dataset_info


def check_index(dataset_path: EPath):
    if not dataset_path.is_dir():
        click.echo(f"{dataset_path} is not a directory, therefore the index will not be checked")
        return

    ok = True

    # Get info file
    info = get_dataset_info(dataset_path)

    click.echo("Checking the index files...")
    shards = info["shard_counts"]
    for shard_file, length in shards.items():
        with TarIndexReader(dataset_path / shard_file) as itar:
            l = len(itar)
            if l - 1 != length:
                ok = False
                print(
                    f"Error in shard {shard_file}: Shard length in Info file {length} != {l - 1} (length in index)"
                )

    return ok


@click.command(name="lint")
@click.argument(
    "path",
    type=click.Path(path_type=EPath),
)
@click.option(
    "--split-parts", default="train,val,test", help="The splits to verify", show_default=True
)
@click.option(
    "--dataset-config", default="dataset.yaml", help="Dataset config file name", show_default=True
)
@click.option(
    "--split-config", default="split.yaml", help="Split config file name", show_default=True
)
@click.option(
    "--parallel", default=1, help="Number of parallel workers", show_default=True, type=int
)
def command(path: EPath, split_parts: str, dataset_config: str, split_config: str, parallel: int):
    """Check energon dataset for errors.

    The PATH should point to the folder with the dataset.
    The dataset must comply with the energon dataset format. See README.md for more details."""

    # Check the tar file index
    if not check_index(path):
        raise click.ClickException("Validation failed with errors, see logs for details.")

    # Check the dataset
    failed = False

    ignore_list = []

    def handler(exc: Exception, key: Optional[str] = None) -> None:
        nonlocal failed
        failed = True
        logging.exception(str(exc))
        if key is not None:
            ignore_list.append(key)

    kwargs = {}
    if dataset_config != "dataset.yaml":
        kwargs["dataset_config"] = dataset_config
    if split_config != "split.yaml":
        kwargs["split_config"] = split_config

    worker_config = WorkerConfig(rank=0, world_size=1, num_workers=parallel)

    for split_part in split_parts.split(","):
        try:
            dataset = get_val_dataset(
                EPath(path),
                split_part=split_part,
                worker_config=worker_config,
                batch_size=1,
                handler=handler,
                **kwargs,
            )
        except EmptyDatasetError:
            click.echo(f"Skipping empty split part {split_part}")
            continue

        try:
            for _ in tqdm.tqdm(get_loader(dataset)):
                pass
        except InterruptedError:
            raise
        except BaseException:
            traceback.print_exc()
            raise click.ClickException("Validation failed with errors, see logs for details.")

    if failed:
        click.echo(
            "The following shards/samples failed (maybe set as dataset.yaml:ignore_list):", err=True
        )
        for item in ignore_list:
            click.echo(f"- {item}", err=True)
        raise click.ClickException("Validation failed with errors, see logs for details.")


if __name__ == "__main__":
    command()



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Callable, Dict, Generic, Iterator, List, Literal, Optional, TypeVar

from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers.base import BaseWrapperDataset

T_sample = TypeVar("T_sample")


def default_get_keys(batch: Any) -> Optional[List[str]]:
    """Default get_keys, which has some heuristics to find the sample keys."""
    if isinstance(batch, list):
        batch = batch[0]
    if (
        hasattr(batch, "__key__")
        and isinstance(batch.__key__, list)
        and all(isinstance(k, str) for k in batch.__key__)
    ):
        return batch.__key__
    elif (
        hasattr(batch, "__keys__")
        and isinstance(batch.__keys__, list)
        and all(isinstance(k, str) for k in batch.__keys__)
    ):
        return batch.__keys__
    elif (
        isinstance(batch, dict)
        and "__key__" in batch
        and all(isinstance(k, str) for k in batch["__key__"])
    ):
        return batch["__key__"]
    elif (
        isinstance(batch, dict)
        and "__keys__" in batch
        and all(isinstance(k, str) for k in batch["__keys__"])
    ):
        return batch["__keys__"]
    elif (
        isinstance(batch, dict)
        and "keys" in batch
        and all(isinstance(k, str) for k in batch["keys"])
    ):
        return batch["keys"]
    return None


class LogSampleDataset(BaseWrapperDataset[T_sample, T_sample], Generic[T_sample]):
    """This dataset logs every yielded sample to the debug logs."""

    get_keys_fn: Callable[[T_sample], Optional[List[str]]]
    mode: Literal["train", "val"]
    _step: int

    _savable_fields = ("_step",)

    def __init__(
        self,
        dataset: SavableDataset[T_sample],
        mode: Literal["train", "val"],
        worker_config: WorkerConfig,
        get_keys_fn: Callable[[T_sample], Optional[List[str]]] = default_get_keys,
    ):
        """Construct the log sample dataset, which logs every yielded sample to the debug logs.

        Args:
            dataset: The input dataset to wrap
        """
        super().__init__(dataset, worker_config=worker_config)
        self.get_keys_fn = get_keys_fn
        self.mode = mode

        self.reset_state_own()

    def reset_state_own(self) -> None:
        self._step = 0

    def __len__(self):
        return len(self.dataset)

    def _log(self, sample: T_sample) -> None:
        if self.worker_config.should_log(level=1):
            log_entry = {
                "t": "yield_batch",
                "r": self.worker_config.rank,
                "w": self.worker_config.global_worker_id(),
                "m": self.mode,
                "idx": self._step,
            }
            keys = self.get_keys_fn(sample)
            if keys is not None:
                log_entry["keys"] = keys

            self.worker_config.worker_log(log_entry)

    def __iter__(self) -> Iterator[T_sample]:
        for sample in self.dataset:
            self._log(sample)
            self._step += 1
            yield sample

    def config(self) -> Dict[str, Any]:
        # Transparent logger, it won't change the samples
        return self.dataset.config()

    def __str__(self):
        return f"LogSampleDataset(mode={self.mode}, get_keys_fn={self.get_keys_fn}, dataset={self.dataset})"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Dict, Literal, Optional, Union

from megatron.energon.dataset_config import get_dataset_from_config
from megatron.energon.edataclass import edataclass
from megatron.energon.epathlib import EPath
from megatron.energon.flavors import BaseCoreDatasetFactory
from megatron.energon.metadataset.loader_interface import (
    DatasetBlendMode,
    DatasetLoaderInterface,
    LoadedDataset,
    LoadedDatasetList,
)
from megatron.energon.worker import WorkerConfig


@edataclass
class DatasetLoader(DatasetLoaderInterface):
    """Loads a dataset from a path."""

    path: Union[str, EPath]
    split_part: Optional[str] = None
    subflavors: Optional[Dict[str, Any]] = None
    shuffle_over_epochs_multiplier: Optional[int] = 1
    dataset_config: str = "dataset.yaml"
    split_config: str = "split.yaml"

    def post_initialize(self, mds_path: Optional[EPath] = None):
        pass

    def get_dataset(
        self,
        *,
        training: bool,
        split_part: Optional[str] = None,
        worker_config: WorkerConfig,
        subflavors: Optional[Dict[str, Any]] = None,
        shuffle_over_epochs: Optional[int] = 1,
        split_config: Optional[str] = None,
        dataset_config: Optional[str] = None,
        **kwargs,
    ) -> BaseCoreDatasetFactory:
        """
        Args:
            training: If true, apply training randomization.
            split_part: Default split part to use.
            worker_config: Worker configuration.
            shuffle_buffer_size: Size of the sample shuffle buffer (before task encoding).
            subflavors: Subflavors to use, might be overridden by inner datasets.
            shuffle_over_epochs: Shuffle the dataset over this many epochs.
            **kwargs: Additional arguments to the dataset constructor.

        Returns:
            The loaded dataset
        """
        if self.split_part is not None:
            split_part = self.split_part
        if split_part is None:
            raise ValueError("Missing split part")
        if self.subflavors is not None:
            subflavors = {**self.subflavors, **(subflavors or {})}
        if split_config is None:
            split_config = self.split_config
        if dataset_config is None:
            dataset_config = self.dataset_config
        return get_dataset_from_config(
            self.path,
            training=training,
            split_part=split_part,
            worker_config=worker_config,
            subflavors=subflavors,
            dataset_config=dataset_config,
            split_config=split_config,
            shuffle_over_epochs=shuffle_over_epochs,
            **kwargs,
        )

    def get_datasets(
        self,
        *,
        training: bool,
        split_part: Union[Literal["train", "val", "test"], str],
        worker_config: WorkerConfig,
        subflavors: Optional[Dict[str, Any]] = None,
        shuffle_over_epochs_multiplier: Optional[int] = 1,
        **kwargs,
    ) -> LoadedDatasetList:
        return LoadedDatasetList(
            blend_mode=DatasetBlendMode.NONE,
            datasets=[
                LoadedDataset(
                    dataset=self.get_dataset(
                        training=training,
                        split_part=split_part,
                        worker_config=worker_config,
                        subflavors=subflavors,
                        shuffle_over_epochs=shuffle_over_epochs_multiplier,
                        **kwargs,
                    ),
                    weight=None,
                )
            ],
        )



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause
from typing import Literal, Optional

import filetype
from bitstring.bits import BitsType
from sortedcontainers import SortedList

from .containers.matroska import parse_matroska
from .containers.mpeg import parse_mpeg
from .containers.probe import parse_probe
from .keyframeinfo import KeyframeInfo


class Fastseek:
    """
    Gathers information from the video container file (e.g. metadata which requires minimal decoding)
    to find keyframes in the video for fast seeking.

    Information is returned in the form of KeyframeInfo structures which can be used by a decoding loop
    to make informed decisions about the best seeking behavior

    Currently supports:
    - MP4/MOV: frames are indexed by number and frame counting can be used to get the exact frame
    - Matroska/WebM: frames are indexed by time and inter-frame duration must be accounted for to get to the right frame

    If your container is not listed above, pass "probe=True" to the constructor, this will use ffmpeg to parse the stream
    without decoding it. Frames will be indexed by number. This is not as fast as using a supported container but is still
    significantly faster than sequential decoding.
    """

    keyframes: dict[int, SortedList[KeyframeInfo]]
    unit: Literal["frames", "pts"]
    mime: str

    def __init__(self, file: BitsType, probe: bool = False) -> None:
        """Initialize the Fastseek object.

        Args:
            file: The video file data as a bitstring BitsType object. This should contain the raw bytes of the video file.
            probe: If True, use ffmpeg to probe the stream without decoding. This is slower but works with any container format.
                   If False (default), attempt to parse the container format directly. Only works with MP4/MOV and Matroska/WebM.

        Raises:
            ValueError: If the file type cannot be determined or if the container format is not supported (when probe=False).
        """
        if probe:
            self.keyframes = parse_probe(file)
            self.unit = "frames"
        else:
            ftype = filetype.guess(file)

            if ftype is None:
                raise ValueError(
                    "Unable to determine file type (hint: try passing probe=True to the Fastseek constructor)"
                )

            self.mime = ftype.mime

            if ftype.mime in ["video/mp4", "video/quicktime"]:
                self.keyframes = parse_mpeg(file)
                self.unit = "frames"
            elif ftype.mime in ["video/x-matroska", "video/webm"]:
                self.keyframes = parse_matroska(file)
                self.unit = "pts"
            else:
                raise ValueError(
                    f"Unsupported container: {ftype.mime} (hint: try passing probe=True to the Fastseek constructor)"
                )

            if len(self.keyframes) == 0:
                raise ValueError(
                    f"The parser for {ftype.mime} was unable to find any streams (hint: try passing probe=True to the Fastseek constructor)"
                )

            if all(len(kf) == 0 for kf in self.keyframes.values()):
                raise ValueError(
                    f"The parser for {ftype.mime} was unable to find any keyframes (hint: try passing probe=True to the Fastseek constructor)"
                )

    def should_seek(self, current: int, target: int, stream: int = 0) -> Optional[KeyframeInfo]:
        """Determine if seeking to a keyframe is necessary to reach the target frame.

        This method helps optimize video seeking by determining whether a seek operation
        is needed to reach the target frame. It returns information about the nearest
        keyframe only if seeking would be beneficial (i.e., if sequential decoding from
        the current position would be less efficient).

        Args:
            current: The current frame number or timestamp (depending on container format)
            target: The desired frame number or timestamp to seek to
            stream: The video stream index to use. Defaults to 0.

        Returns:
            Information about the nearest keyframe if seeking would be beneficial,
            or None if sequential decoding from current position is more efficient.
            The KeyframeInfo contains the keyframe's position and timing information.

        Note:
            The units for current and target depend on the container format:
            - For MP4/MOV: frame numbers (count-based)
            - For Matroska/WebM: timestamps (time-based)
        """
        nearest_iframe: KeyframeInfo = self.nearest_keyframe(target, stream)
        return (
            nearest_iframe
            if (current < nearest_iframe.index <= target) or (target < current)
            else None
        )

    def nearest_keyframe(self, target: int, stream: int = 0) -> KeyframeInfo:
        """Find the nearest keyframe that comes before the target frame.

        This method performs a binary search to find the keyframe that is closest to,
        but not after, the target frame position. This is useful for determining the
        optimal starting point for decoding to reach a specific frame.

        Args:
            target: The target frame number or timestamp to find the nearest keyframe for.
                The unit (frame count or timestamp) depends on the container format.
            stream: The video stream index to use. Defaults to 0.
                Used when the container has multiple video streams.

        Returns:
            Information about the nearest keyframe before the target position.
            Contains details like the keyframe's position, timestamp, and file offset.

        Note:
            The implementation currently uses a list-based approach for stream selection
            as some video containers don't report track IDs correctly. This is a temporary
            workaround and may be updated in the future.
        """

        if stream >= len(self.keyframes):
            raise ValueError(f"No stream with index {stream}")

        stream_id = list(self.keyframes.keys())[stream]

        if len(self.keyframes[stream_id]) == 0:
            raise ValueError(f"No keyframes found for stream {stream}")

        nearest_iframe_to_target_index: int = self.keyframes[stream_id].bisect_left(target) - 1
        return self.keyframes[stream_id][max(0, nearest_iframe_to_target_index)]



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from importlib import import_module
from typing import Type

import braceexpand
import click

import megatron.energon
from megatron.energon.epathlib import EPath
from megatron.energon.eyaml import load_yaml, load_yaml_json
from megatron.energon.flavors.base_dataset import BaseCoreDatasetFactory
from megatron.energon.flavors.webdataset import MAIN_FOLDER_NAME
from megatron.energon.flavors.webdataset.metadata import get_dataset_info

fmt = """
* {dataset_name}: `{path}` ({samples_count:,} samples, {samples_size} in {shards_count} shards)
  * Created with energon version: {energon_version}
  * Sample Type: {{py:class}}`{sample_name} <{sample_fullname}>`  
  * Default Splits:
{splits_str}
"""

split_fmt = """    * `{split_name}`: {split_ratio:.0f}%, {split_samples_count:,} samples in {split_shards_count} shards
"""


def fmt_size(size: int) -> str:
    keys = ["B", "KiB", "MiB", "GiB", "TiB"]
    for key in keys:
        if size < 1024:
            return f"{size:.2f} {key}"
        size /= 1024
    return f"{size:.2f} PiB"


@click.command(name="info")
@click.argument(
    "path",
    type=click.Path(file_okay=False, dir_okay=True, path_type=EPath),
)
@click.option(
    "--split-config", default="split.yaml", help="Split config file name", show_default=True
)
@click.option(
    "--dataset-config", default="dataset.yaml", help="Dataset config file name", show_default=True
)
def command(
    path: EPath,
    split_config: str,
    dataset_config: str,
):
    """
    Get summarizing information about a dataset.
    """

    ds_config = load_yaml((path / MAIN_FOLDER_NAME / dataset_config).read_bytes())
    info_config = get_dataset_info(path)
    split_config_obj = load_yaml_json(path / MAIN_FOLDER_NAME / split_config)

    ds_energon_version = info_config.get("energon_version", "unknown")
    samples_count = sum(info_config["shard_counts"].values())
    dict_sample_type = ds_config["sample_type"]
    sample_module = import_module(dict_sample_type["__module__"])

    sample_cls: Type[BaseCoreDatasetFactory] = getattr(sample_module, dict_sample_type["__class__"])
    sample_module = sample_cls.__module__
    if (
        sample_module.startswith("megatron.energon")
        and getattr(megatron.energon, dict_sample_type["__class__"], None) == sample_cls
    ):
        sample_module = "megatron.energon"
    sample_name = sample_cls.__name__
    sample_fullname = sample_module + "." + sample_name

    def srt_key(pair):
        try:
            return ("train", "val", "test").index(pair[0])
        except ValueError:
            return 3

    # Brace expand all the split part files
    expanded_split_parts = {}
    for split_name, split_parts in split_config_obj["split_parts"].items():
        expanded_split_parts[split_name] = []
        for split_part in split_parts:
            for name in braceexpand.braceexpand(split_part):
                expanded_split_parts[split_name].append(name)

    splits_str = "".join(
        split_fmt.format(
            split_name=split_name,
            split_ratio=round(
                100
                * sum(info_config["shard_counts"][shard] for shard in split_parts)
                / samples_count,
                2,
            ),
            split_samples_count=sum(info_config["shard_counts"][shard] for shard in split_parts),
            split_shards_count=len(split_parts),
        )
        for split_name, split_parts in sorted(expanded_split_parts.items(), key=srt_key)
    )
    print(
        fmt.format(
            dataset_name=path.name,
            path=str(path),
            samples_count=samples_count,
            samples_size=fmt_size(
                sum((path / split_name).size() for split_name in info_config["shard_counts"].keys())
            ),
            shards_count=len(info_config["shard_counts"]),
            sample_name=sample_name,
            sample_fullname=sample_fullname,
            splits_str=splits_str,
            energon_version=ds_energon_version,
        )
    )



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import functools
from dataclasses import dataclass
from typing import Callable, Generic, Optional, TypeVar, Union, overload

from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.crude import CrudeSample
from megatron.energon.source_info import SourceInfo

T_sample = TypeVar("T_sample", bound=Sample, covariant=True)

F = TypeVar("F", bound=Callable[..., Sample])


@overload
def cooker(
    fn: None = None,
) -> Callable[[F], F]: ...


@overload
def cooker(
    *,
    need_cache: bool = False,
    need_primary: bool = False,
) -> Callable[[F], F]: ...


def cooker(
    fn: Optional[F] = None,
    *,
    need_cache: bool = False,
    need_primary: bool = False,
) -> Union[
    F,
    Callable[[F], F],
]:
    """Decorator to mark a function as a cooker, optionally enabling cache and primary dataset
    arguments."""
    if fn is None:
        return functools.partial(
            cooker,
            need_cache=need_cache,
            need_primary=need_primary,
        )

    @functools.wraps(fn)
    def fn_wrapper(*args, **kwargs):
        return fn(*args, **kwargs)

    setattr(fn_wrapper, "__cooker_need_cache__", need_cache)
    setattr(fn_wrapper, "__cooker_need_primary__", need_primary)
    return fn_wrapper


def get_cooker_need_cache(fn: Callable[..., T_sample]) -> bool:
    """Get whether a function is a cooker."""
    return getattr(fn, "__cooker_need_cache__", False)


def get_cooker_need_primary(fn: Callable[..., T_sample]) -> bool:
    """Get whether a function is a cooker."""
    return getattr(fn, "__cooker_need_primary__", False)


@dataclass
class Cooker(Generic[T_sample]):
    """A cooker transforms a crude sample (simple dict) into a specific sample type inheriting
    from `Sample`.
    The `cook` method performs the transformation, the other fields are used to select the
    samples which this cooker can transform. If no filters are provided, the cooker will transform
    any `CrudeSample`.
    """

    #: The callable that performs the cooking (i.e. loading / transforming the crude sample).
    # Signature is:
    # `(/, raw_sample: dict, *, primary?: RandomAccessDataset, **aux: RandomAccessDataset, cache?: Cache) -> Sample`.
    # `primary` is passed only if want_primary_random_access is true.
    # `cache` is passed only if want_cache is true.
    cook: Callable[..., T_sample]

    #: The subflavors to be present in the sample to be cooked by this cooker. All keys and values
    # must match.
    has_subflavors: Optional[dict] = None

    @property
    def need_primary(self) -> bool:
        return get_cooker_need_primary(self.cook)

    @property
    def need_cache(self) -> bool:
        return get_cooker_need_cache(self.cook)

    def is_match(self, crude_sample: CrudeSample) -> bool:
        if self.has_subflavors is not None:
            # Checks if the dict entries provided as a filter all match
            # the ones in the sample. The sample may have additional entries.
            for k, v in self.has_subflavors.items():
                if (
                    k not in crude_sample["__subflavors__"]
                    or crude_sample["__subflavors__"][k] != v
                ):
                    return False

        return True


def basic_sample_keys(
    crude_sample: dict, additional_source_info: tuple[SourceInfo, ...] = ()
) -> dict:
    """A convenience helper to extract the basic keys from a crude sample,
    which you will always need to forward to the cooked sample."""

    res = {
        field.name: crude_sample[field.name]
        for field in dataclasses.fields(Sample)
        if field.name in crude_sample
    }

    if additional_source_info:
        res["__sources__"] = (*crude_sample["__sources__"], *additional_source_info)
    return res



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Dict, Generic, Optional, TypeVar

from megatron.energon.epathlib import EPath
from megatron.energon.flavors.webdataset.default_generic_webdataset import (
    DefaultGenericWebdatasetFactory,
)
from megatron.energon.flavors.webdataset.sample_decoder import SampleDecoder
from megatron.energon.flavors.webdataset.structs import FilteredSample

T_sample = TypeVar("T_sample", covariant=True)


class DefaultDecoderWebdatasetFactory(DefaultGenericWebdatasetFactory[T_sample], Generic[T_sample]):
    """
    Extends the default webdataset loading with decoding of contained files, such as images, videos or nested
    containers.
    """

    # The webdataset decoder function, if to be applied
    _decoder: Optional[SampleDecoder]

    def __init__(
        self,
        path: EPath,
        *,
        decoder: Optional[SampleDecoder] = SampleDecoder(),
        **kwargs,
    ):
        """
        Factory for the webdataset sample loader including the decoder.

        Args:
            path: Path to the dataset (passed to parent)
            decoder: If provided, use this decoder, otherwise just load raw bytes.
            **kwargs: Args passed to parent constructor
        """
        self._decoder = decoder
        super().__init__(path, **kwargs)

    def load_sample(self, sample: FilteredSample) -> T_sample:
        if self._decoder is not None:
            sample = self._decoder(sample)
        return super().load_sample(sample)

    def config(self) -> Dict[str, Any]:
        return dict(
            **super().config(),
            **(self._decoder.config() if self._decoder is not None else {}),
        )



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Callable, Generic, Optional, Tuple, TypeVar, Union

from megatron.energon.errors import SYSTEM_EXCEPTIONS, FatalSampleError
from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.webdataset.structs import FilteredSample, reraise_exception
from megatron.energon.source_info import SourceInfo

T_sample = TypeVar("T_sample", covariant=True)


class ErrorHandler(Generic[T_sample]):
    handler: Callable[[Exception, Optional[str], Optional[list[SourceInfo]]], None] = (
        reraise_exception
    )

    def sample_error_handler(
        self, e: Exception, sample_key: Optional[str], sources: Optional[list[SourceInfo]] = None
    ):
        if isinstance(e, SYSTEM_EXCEPTIONS):
            raise FatalSampleError(f"Error in sample {sample_key!r}: {e}") from e

        self.handler(e, sample_key, sources)

    def error_handler(
        self,
        e: Exception,
        sample: Union[
            T_sample,
            dict,
            FilteredSample,
            None,
            Tuple[Union[T_sample, dict, FilteredSample, None], ...],
        ],
    ):
        if isinstance(sample, dict):
            key = sample.get("__key__")
            sources = sample.get("__sources__")
        elif isinstance(sample, list):
            if isinstance(sample[0], dict):
                key = ",".join("None" if s is None else s.get("__key__") for s in sample)
                sources = [src for s in sample for src in s.get("__sources__", ())]
            elif isinstance(sample[0], Sample):
                key = ",".join("None" if s is None else s.__key__ for s in sample)
                sources = [src for s in sample for src in s.__sources__]
            else:
                key = None
                sources = None
        elif isinstance(sample, Sample):
            key = sample.__key__
            sources = sample.__sources__
        else:
            key = None
            sources = None
        self.sample_error_handler(e, key, sources)



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import configparser
import os
import shutil
from pathlib import Path
from typing import Dict, Optional

from megatron.energon.edataclass import edataclass


@edataclass
class ConfigEntry:
    name: str
    type: str
    provider: Optional[str]
    access_key_id: Optional[str]
    secret_access_key: Optional[str]
    region: Optional[str]
    endpoint: Optional[str]


def find_executable_path(executable_name):
    """Find the path of an executable in the PATH environment variable. Returns None if not found."""

    executable_path = shutil.which(executable_name)
    if executable_path:
        return Path(executable_path)
    return None


def get_rclone_config_path() -> Optional[Path]:
    # First check if rclone executable is in PATH, if yes, check if rclone.conf is in the same directory
    rclone_exe_path = find_executable_path("rclone")
    if rclone_exe_path is not None and rclone_exe_path.is_file():
        rclone_config_path = rclone_exe_path.with_name("rclone.conf")
        if rclone_config_path.is_file():
            return rclone_config_path

    # As a second option check the XDG_CONFIG_HOME environment variable, if it is set, check for rclone/rclone.conf in that directory
    xdg_config_home = os.environ.get("XDG_CONFIG_HOME")
    if xdg_config_home and Path(xdg_config_home).is_dir():
        rclone_config_path = Path(xdg_config_home) / "rclone" / "rclone.conf"
        if rclone_config_path.is_file():
            return rclone_config_path

    # As a third option check the default location ~/.config/rclone/rclone.conf
    rclone_config_path = Path.home() / ".config" / "rclone" / "rclone.conf"
    if rclone_config_path.is_file():
        return rclone_config_path

    # Last option is to check the legacy location ~/.rclone.conf
    legacy_config_path = Path.home() / ".rclone.conf"
    if legacy_config_path.is_file():
        return legacy_config_path

    return None


def read_rclone_config_at_path(config_path: Path) -> Dict[str, ConfigEntry]:
    """Reads the config file and returns a dictionary with the config entries."""

    config = configparser.ConfigParser()
    config.read(config_path)

    config_entries = {}
    for section in config.sections():
        entry = ConfigEntry(
            name=section,
            type=config[section].get("type"),
            provider=config[section].get("provider"),
            access_key_id=config[section].get("access_key_id"),
            secret_access_key=config[section].get("secret_access_key"),
            region=config[section].get("region"),
            endpoint=config[section].get("endpoint"),
        )
        config_entries[section] = entry

    return config_entries


def read_rclone_config() -> Dict[str, ConfigEntry]:
    config_path = get_rclone_config_path()
    if config_path is None:
        raise FileNotFoundError("Could not find rclone configuration file.")
    return read_rclone_config_at_path(config_path)



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Any, Generator, Generic, Iterable, Optional, Tuple, Type, TypeVar, Union

from megatron.energon.flavors.base_dataset import (
    FlexState,
    Sample,
    SavableDataset,
    add_sample_restore_key,
)
from megatron.energon.savable import Savable
from megatron.energon.worker import WorkerConfig

T = TypeVar("T")
T_sample = TypeVar("T_sample", covariant=True)
T_sample_out = TypeVar("T_sample_out", covariant=True)
T_sample_in = TypeVar("T_sample_in", covariant=True)


class BaseWrapperDataset(SavableDataset[T_sample_out], Generic[T_sample_in, T_sample_out], ABC):
    """Base class for dataset wrappers. All dataset wrappers should derive from this. A dataset
    wrapper takes one dataset and modifies its samples to make a new dataset. This can be for
    shuffling samples or applying custom functions to the data. Some wrappers only modify the
    length of the dataset or how it's repeated."""

    datasets: Tuple[SavableDataset[T_sample_in], ...]

    def __init__(
        self,
        datasets: Union[SavableDataset[T_sample_in], Iterable[SavableDataset[T_sample_in]]],
        *,
        worker_config: WorkerConfig,
    ):
        super().__init__(worker_config=worker_config)

        if isinstance(datasets, SavableDataset):
            self.datasets = (datasets,)
        else:
            self.datasets = tuple(datasets)

        for d in self.datasets:
            # Check that the dataset worker configs are the same as the wrapper worker config
            assert d.worker_config == self.worker_config, (
                "Dataset and wrapper worker configs must match."
            )

    @property
    def dataset(self) -> SavableDataset:
        """Convenience property, if only one dataset is wrapped."""

        assert len(self.datasets) == 1
        return self.datasets[0]

    def can_restore_sample(self) -> bool:
        return all(ds.can_restore_sample() for ds in self.datasets)

    def assert_can_restore(self) -> None:
        for ds in self.datasets:
            ds.assert_can_restore()

    def worker_has_samples(self) -> bool:
        return any(ds.worker_has_samples() for ds in self.datasets)

    def _find_wrapped_dataset(self, cls: Type[SavableDataset]) -> Optional[SavableDataset]:
        """Find the outermost dataset wrapped in this dataset that is of type cls."""

        for ds in self.datasets:
            if isinstance(ds, cls):
                return ds
            elif isinstance(ds, BaseWrapperDataset):
                res = ds._find_wrapped_dataset(cls)
                if res is not None:
                    return res
        return None

    def restore_sample(self, restore_key: Tuple[Union[str, int, tuple], ...]) -> T_sample_out:
        if len(self.datasets) == 1:
            return self.datasets[0].restore_sample(restore_key)
        else:
            id, ds_idx = restore_key[:2]
            assert id == type(self).__name__
            restore_key = restore_key[2:]
            assert isinstance(ds_idx, int)
            return add_sample_restore_key(
                self.datasets[ds_idx].restore_sample(restore_key),
                ds_idx,
                src=self,
            )

    def save_state(self) -> FlexState:
        own_state = super().save_state()

        return FlexState(datasets=[ds.save_state() for ds in self.datasets], **own_state)

    def restore_state(self, state: FlexState) -> None:
        assert len(self.datasets) == len(state["datasets"])
        for dataset, dstate in zip(self.datasets, state["datasets"]):
            dataset.restore_state(dstate)

        super().restore_state(state)

    def reset_state_deep(self) -> None:
        """Resets the state of the inner datasets and then the own state."""

        for ds in self.datasets:
            if isinstance(ds, BaseWrapperDataset):
                ds.reset_state_deep()
            else:
                ds.reset_state_own()

        self.reset_state_own()

    @abstractmethod
    def reset_state_own(self) -> None:
        """Resets the state of the dataset, excl. the inner datasets."""
        ...


class SampleIndex(Savable):
    """A simple class to hold the sample index for one worker."""

    worker_config: WorkerConfig
    current_idx: int

    actives = 0

    def __init__(self, worker_config: WorkerConfig, *, src: Any) -> None:
        self.worker_config = worker_config
        self.current_idx = 0
        self.src = src

    def get_next(self) -> int:
        res = self.current_idx
        self.current_idx += 1
        return res

    @contextmanager
    def ctx(self, sample_idx: Optional[int] = None):
        if sample_idx is None:
            sample_idx = self.get_next()
        assert WorkerConfig.active_worker_config is not None
        WorkerConfig.active_worker_config.worker_push_sample_index(sample_idx)
        # print("  " * SampleIndex.actives + f"Activated from {type(self.src).__name__}({id(self.src)}) {sample_idx} -> {WorkerConfig.active_worker_config._sample_index_stack}")
        SampleIndex.actives += 1
        try:
            yield sample_idx
        finally:
            assert WorkerConfig.active_worker_config is not None
            popped = WorkerConfig.active_worker_config.worker_pop_sample_index()
            SampleIndex.actives -= 1
            # print("  " * SampleIndex.actives + f"Deactivate from {type(self.src).__name__}({id(self.src)}) {sample_idx} -> {WorkerConfig.active_worker_config._sample_index_stack}")
            assert popped == sample_idx, f"Expected {sample_idx}, got {popped}"

    def iter_ctx(
        self,
        it: Iterable[T_sample],
        sample_idx: Optional[int] = None,
    ) -> Generator[Tuple[int, T_sample], None, None]:
        it = iter(it)
        try:
            while True:
                try:
                    with self.ctx(sample_idx) as res_sample_idx:
                        x = next(it)
                    yield res_sample_idx, x
                except StopIteration:
                    break
        finally:
            if hasattr(it, "close"):
                it.close()

    def save_state(self) -> int:
        return self.current_idx

    def restore_state(self, state: Optional[int]) -> None:
        if state is None:
            self.current_idx = 0
        else:
            self.current_idx = state


def get_sample_restore_key(sample: Any) -> Optional[Union[str, int]]:
    """Gets the restore key from an arbitrary sample."""
    if isinstance(sample, Sample) or hasattr(sample, "__restore_key__"):
        return sample.__restore_key__
    elif isinstance(sample, dict) and "__restore_key__" in sample:
        return sample["__restore_key__"]
    else:
        return None



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import hashlib
import os
import pickle
import random
import string
import tempfile
import threading
from concurrent.futures import Future, ThreadPoolExecutor
from pathlib import Path
from typing import Any, Dict, Literal, Optional, Tuple, TypeVar

from megatron.energon.cache.base import CachePool, FileStore, Lazy
from megatron.energon.cache.file_store import DecodeFileStore
from megatron.energon.edataclass import edataclass
from megatron.energon.fork_hook import ForkMixin
from megatron.energon.source_info import SourceInfo, add_source_info

T = TypeVar("T")


@edataclass
class FileCacheLazy(Lazy[T]):
    """
    Represents a reference to a background prefetch.
    """

    # The cache pool that we're using.
    pool: "FileStoreCachePool"
    # The entry in the cache pool that we're using.
    entry: "_PendingTask"

    # If get() was called, this will be the data (uncached).
    _data: Optional[tuple[T, SourceInfo]] = None

    def get(self, sample: Any = None) -> T:
        """
        Returns the data and adds the source info to the sample.
        If the background job hasn't started, we cancel it,
        do a direct read, and remove ourselves from the pool's references.
        Otherwise, we wait for the job to finish, read from cache, and remove ourselves.
        """
        if self._data is not None:
            return self._data
        self._data = self.pool._get_data(self.ds, self.fname, self.entry)
        assert self._data is not None
        add_source_info(sample, self._data[1])
        return self._data[0]

    def __hash__(self) -> int:
        """Allows usage in sets and dicts as key."""
        return hash((id(self.ds), self.fname))

    def __eq__(self, other: Any) -> bool:
        """Allows usage in sets and dicts as key. Compares the data source and the filename."""
        if not isinstance(other, Lazy):
            return False
        return self.ds is other.ds and self.fname == other.fname

    def __del__(self):
        if self._data is None:
            with self.pool._lock:
                # Data was never fetched, still decrement refcount to delete the cache entry
                self.pool._decrement_refcount_and_cleanup((self.ds.get_path(), self.fname))


@edataclass
class _PendingTask:
    """Dataclass for storing a pending background task"""

    # The dataset that we're caching.
    ds: FileStore
    # The file name that we're caching.
    fname: str
    # The future for the background task that sends the data to the cache.
    send_to_cache_future: Future
    # The number of references to the cache entry.
    refcount: int = 1
    # The size of the data to be cached.
    data_size: int = 0
    # Whether the data is required now, i.e. a reading thread is waiting for it.
    require_data_now: bool = False
    # The path to the cache file.
    cache_path: Optional[Path] = None
    # The source info for the data.
    source_info: Optional[SourceInfo] = None


class FileStoreCachePool(CachePool, ForkMixin):
    """
    Manages a thread pool to pre-fetch data onto an SSD cache.
    Each (ds, fname) has one Future (one read). Multiple requests
    share that same future. We track usage with a refcount.

    To avoid multi-process collisions, we generate a random subfolder
    for each instance.
    """

    cache_dir: Path
    max_cache_size: int
    max_cache_count: int
    current_cache_size: int
    current_cache_count: int
    method: Literal["raw", "pickle"]

    # Thread pool for out-caching tasks
    _worker_pool: Optional[ThreadPoolExecutor] = None
    # (ds.path, fname) -> PendingTask
    _pending_tasks: Dict[Tuple[str, str], _PendingTask]

    # Lock for all shared structures
    _lock: threading.Lock
    # Condition variable to signal when cache space is available
    _cache_space_available: threading.Condition
    # Whether the pool is shutting down
    _shutting_down: bool = False

    def __init__(
        self,
        *,
        parent_cache_dir: Optional[Path] = None,
        num_workers: int = 8,
        max_cache_size_gbytes: float = 1024,
        max_cache_count: int = 10_000_000,
        method: Literal["raw", "pickle"] = "raw",
    ):
        """
        Initialize the cache pool.

        Args:
            parent_cache_dir: The parent directory for the cache.
            num_workers: The number of worker threads to use for copying the data to the cache for lazy loading.
            max_cache_size_gbytes: The maximum size of the cache in gigabytes. If the cache exceeds this size,
                the prefetching will wait until the cache is below this size.
            max_cache_count: The maximum number of files in the cache. If the cache exceeds this number,
                the prefetching will wait until the cache is below this number.
            method: The method to use for caching. "raw" store the non-decoded raw data. "pickle": first decode the data
                and then store the pickled data.
        """
        super().__init__()
        # If no parent directory is given, create a temp directory
        if parent_cache_dir is None:
            parent_cache_dir = Path(tempfile.gettempdir())
        self.parent_cache_dir = parent_cache_dir
        self.num_workers = num_workers

        # Initialize the cache pool (process volatile fields)
        self.__after_fork__(initial=True)

        self.method = method

        # We'll store _pending_tasks in the form:
        #   (ds.path, fname) -> PendingTask
        self._pending_tasks = {}

        # Cache size management
        self.max_cache_size = int(max_cache_size_gbytes * (1024**3))
        self.max_cache_count = max_cache_count
        self.current_cache_size = 0
        self.current_cache_count = 0

        # A lock to protect all shared structures
        self._lock = threading.Lock()

        # Condition variable to signal when cache space is available
        self._cache_space_available = threading.Condition(self._lock)

    def get(self, ds: FileStore, fname: str, sample: Any = None) -> Any:
        """
        Synchronous read from the dataset (no cache usage).
        """
        return ds.get(fname, sample)

    def _get_data(self, ds: FileStore, fname: str, entry: _PendingTask) -> tuple[Any, SourceInfo]:
        """
        Get the data for a given file from the cache and purge cache if no references are left.

        * If the cache-out is complete, read from cache.
        * If the cache-out is currently prefetching the data to local storage, wait until it's done.
        * If the cache-out job is waiting for space, skip the cache and do a direct read.
        * If the cache-out job is queued for caching, cancel and do a direct read.
        * If the cache-out job failed, raise through and keep for other references.
        * If the cache-out job is cancelled, requeue if there are other references waiting for it.
        """
        result: tuple[Any, SourceInfo]
        with self._lock:
            try:
                # Attempt to cancel if the job hasn't started
                if entry.send_to_cache_future.cancel():
                    was_cached = False
                    try:
                        # Cancelled => job never ran. We'll do a direct read.
                        result = ds[fname]
                    finally:
                        # Decrement refcount
                        self._decrement_refcount_and_cleanup(key=(ds.get_path(), fname))
                else:
                    # Future is already running or done.
                    # Release the lock so the background job can proceed,
                    # then reacquire it after waiting. Otherwise we might block the worker.
                    entry.require_data_now = True
                    self._cache_space_available.notify_all()
                    self._lock.release()

                    # If the job failed, let's keep the exception for other references.
                    was_cached = True

                    try:
                        # Can raise exception if job failed
                        was_cached = entry.send_to_cache_future.result()

                        if was_cached:
                            # The job is complete; read from cache
                            result = self._read_from_cache(entry)
                        else:
                            # The job failed, so we'll do a direct decode
                            result = ds[fname]
                    finally:
                        self._lock.acquire()
                        entry.require_data_now = False

                        # Decrement refcount
                        self._decrement_refcount_and_cleanup(key=(ds.get_path(), fname))
            finally:
                if entry.refcount > 0 and not was_cached:
                    # TODO: Could write to cache here, data is already fetched.
                    # Write the result to the cache
                    # Requeue the job, there is another reference to the cache entry
                    entry.send_to_cache_future = self._worker_pool.submit(
                        self._cache_out_task, ds, fname, entry
                    )

            return result

    def _cache_out_task(self, ds: FileStore, fname: str, entry: _PendingTask) -> bool:
        with self._lock:
            if self._shutting_down:
                return False

        # Perform the data read
        if self.method == "raw":
            if isinstance(ds, DecodeFileStore):
                data, entry.source_info = ds.inner_reader[fname]
            else:
                data, entry.source_info = ds[fname]
        elif self.method == "pickle":
            data, entry.source_info = ds[fname]
            data = pickle.dumps(data)
        else:
            raise ValueError(f"Invalid method: {self.method}")

        # Wait until there's enough space in the cache
        with self._lock:
            entry.data_size = file_size = len(data)

            while (
                self.current_cache_count + 1 > self.max_cache_count
                or self.current_cache_size + entry.data_size > self.max_cache_size
            ):
                # Release the lock and wait for notification
                self._cache_space_available.wait()
                if entry.require_data_now or self._shutting_down:
                    # At least one reference requires the data now, stop waiting for space and exit immediately
                    return False

            # Reserve the space
            self.current_cache_size += file_size
            self.current_cache_count += 1

            if self._shutting_down or entry.refcount <= 0:
                # No more references to this background job, don't write to cache
                return False

        try:
            assert entry.cache_path is None, (
                f"cache_path should be None, but is {entry.cache_path!r}"
            )
            # Write to cache
            cache_path = self._make_cache_path(ds, fname)
            self._write_to_cache(cache_path, data)
        except:
            with self._lock:
                # Revert the space reservation
                self.current_cache_size -= file_size
                self.current_cache_count -= 1
                self._cache_space_available.notify_all()
            raise
        else:
            with self._lock:
                entry.cache_path = cache_path
            # print(
            #     f"FSCP r={torch.distributed.get_rank()}, pid={os.getpid()}: Wrote to cache {cache_path} (rc={entry.refcount}, size={file_size}, name={fname})\n",
            #     end="",
            # )

        # Data is cached now, return True
        return True

    def get_lazy(self, ds: FileStore, fname: str) -> FileCacheLazy:
        """
        Schedule a background pre-fetch. If multiple calls come in for the same (ds, fname),
        they'll share the same Future and increment reference counts.
        """
        key = (ds.get_path(), fname)
        with self._lock:
            if self._shutting_down:
                raise RuntimeError("Cache pool is already shutting down")
            entry = self._pending_tasks.get(key)
            if entry:
                # Already have a background task for this (ds, fname)
                entry.refcount += 1
            else:
                # Create a new background task
                entry = _PendingTask(
                    ds=ds,
                    fname=fname,
                    send_to_cache_future=None,
                )
                self._pending_tasks[key] = entry

                entry.send_to_cache_future = self._worker_pool.submit(
                    self._cache_out_task, ds, fname, entry
                )

        return FileCacheLazy(ds=ds, fname=fname, pool=self, entry=entry)

    def close(self) -> None:
        """
        Shutdown the pool, wait for tasks, and clear our structures.
        """
        with self._lock:
            self._shutting_down = True
            for entry in self._pending_tasks.values():
                entry.send_to_cache_future.cancel()
            self._cache_space_available.notify_all()
        self._worker_pool.shutdown(wait=True)
        with self._lock:
            self._pending_tasks.clear()

    def _decrement_refcount_and_cleanup(self, key: Tuple[FileStore, str]) -> None:
        """
        Decrement the reference count in `_pending_tasks`.
        If it hits zero, remove the entry. Optionally remove the file if so.
        Assumes the caller holds `self._lock`.
        """
        entry = self._pending_tasks.get(key)
        if not entry:
            # Already cleaned up
            return

        entry.refcount -= 1
        if entry.refcount <= 0:
            # No more references to this background job
            del self._pending_tasks[key]

            self._remove_cached_file(entry)

            assert entry.refcount == 0, f"refcount should be 0: {entry.refcount}"

    def _make_cache_path(self, ds: FileStore, fname: str) -> Path:
        # This is safe, because the parent cache dir is unique per instance.
        ds_hash = hashlib.md5(ds.get_path().encode("utf-8")).hexdigest()
        fn_hash = hashlib.md5(fname.encode("utf-8")).hexdigest()
        # ds_hash = str(ds.get_path()).replace("/", "_")
        # fn_hash = fname.replace("/", "_")
        return self.cache_dir / f"{ds_hash}_{fn_hash}"

    def _write_to_cache(self, path: Path, data: bytes) -> None:
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "wb") as f:
            f.write(data)

    def _read_from_cache(self, entry: _PendingTask) -> tuple[Any, SourceInfo]:
        assert entry.source_info is not None, "source_info should have been set"
        with open(entry.cache_path, "rb") as f:
            if self.method == "raw":
                raw = f.read()
                if isinstance(entry.ds, DecodeFileStore):
                    return entry.ds.decoder.decode(entry.fname, raw), entry.source_info
                else:
                    return raw, entry.source_info
            else:
                return pickle.load(f), entry.source_info

    def _remove_cached_file(self, entry: _PendingTask) -> None:
        """
        Removes a file from disk and updates size counters.
        Assumes the caller holds `self._lock`.
        """
        if entry.cache_path is None:
            return
        if not entry.cache_path.exists():
            return

        try:
            # print(
            #     f"FSCP r={torch.distributed.get_rank()}, pid={os.getpid()}: Removing cached file {entry.cache_path} (rc={entry.refcount})\n",
            #     end="",
            # )
            entry.cache_path.unlink()
        except OSError:
            pass
        entry.cache_path = None

        if entry.data_size > 0:
            self.current_cache_size -= entry.data_size
            self.current_cache_count -= 1
            # Notify waiting threads that space is now available
            self._cache_space_available.notify_all()

    def __before_fork__(self):
        # Ensure the worker pool is shutdown before the fork
        assert len(self._pending_tasks) == 0, "Pending tasks should be empty before fork"
        # print(
        #     f"FSCP r={torch.distributed.get_rank()}, pid={os.getpid()}: Before fork for oid={id(self)} random_suffix={self.cache_dir.name!r}\n",
        #     end="",
        # )
        self._worker_pool.shutdown(wait=True)
        self._worker_pool = None

    def __after_in_child_fork__(self):
        self.__after_fork__()

    def __after_in_parent_fork__(self):
        self.__after_fork__()

    def __after_fork__(self, initial: bool = False):
        random_suffix = "".join(
            random.Random(os.getpid() ^ random.randint(0, 2**32)).choices(
                string.ascii_lowercase + string.digits, k=16
            )
        )
        assert self._worker_pool is None
        self._worker_pool = ThreadPoolExecutor(
            max_workers=self.num_workers, thread_name_prefix="CacheWorker"
        )
        # Create a random subdirectory name to avoid collisions with other processes
        # As the global random generator is cloned across processes, we need to use a process-specific seed
        self.cache_dir = (self.parent_cache_dir / f"cache_{random_suffix}").resolve()
        self.cache_dir.mkdir(parents=True, exist_ok=True)
        # if initial:
        #     print(
        #         f"FSCP r={torch.distributed.get_rank()}, pid={os.getpid()}: Init oid={id(self)} random_suffix={random_suffix!r}\n",
        #         end="",
        #     )
        # else:
        #     print(
        #         f"FSCP r={torch.distributed.get_rank()}, pid={os.getpid()}: After fork for pid={os.getpid()} oid={id(self)} random_suffix={random_suffix!r}\n",
        #         end="",
        #     )

    def __str__(self):
        return f"FileStoreCachePool(cache_dir={self.cache_dir}, max_cache_size={self.max_cache_size}, max_cache_count={self.max_cache_count}, method={self.method}, current_cache_size={self.current_cache_size}, current_cache_count={self.current_cache_count})"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC, abstractmethod
from bisect import bisect_right
from typing import (
    Callable,
    Dict,
    Generator,
    Generic,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from megatron.energon.epathlib import EPath
from megatron.energon.flavors.webdataset.config import skip_meta_re, split_name_re
from megatron.energon.flavors.webdataset.indexing import JoinIndexReader, SqliteIndexReader
from megatron.energon.flavors.webdataset.itar import (
    CachedItarOffsetReader,
    ITarFile,
    ITarSamplePointer,
)
from megatron.energon.flavors.webdataset.metadata import get_info_shard_files
from megatron.energon.flavors.webdataset.structs import FilteredSample, ShardInfo
from megatron.energon.source_info import SourceInfo

T_index = TypeVar("T_index", covariant=False)


class ITarReader(ABC, Generic[T_index]):
    """
    An abstract base class for reading a sequence of tar files containing samples.

    The common usage patterns and random-access interfaces are provided here.

    Args:
        base_path: The base path for the relative tar_filepaths.
        tar_filenames: The list of tar file names.
        tar_filepaths: The corresponding list of full paths to the tar files.
        part_filter: An optional filter function to select parts of the samples.
        itar_cache_size: The number of tar readers to keep open at the same time.
    """

    base_path: EPath
    tar_filenames: List[str]
    tar_filepaths: List[EPath]
    part_filter: Optional[Callable[[str], bool]]
    itar_files_cache: Dict[int, ITarFile]
    sample_filter: Optional[Callable[[str], bool]]

    def __init__(
        self,
        base_path: EPath,
        tar_filenames: List[str],
        tar_filepaths: List[EPath],
        part_filter: Optional[Callable[[str], bool]] = None,
        itar_cache_size: int = 5,
        sample_filter: Optional[Callable[[str], bool]] = None,
    ):
        assert len(tar_filenames) == len(tar_filepaths), (
            f"tar_filenames length ({len(tar_filenames)}) does not match "
            f"tar_filepaths length ({len(tar_filepaths)})"
        )
        self.base_path = base_path
        self.tar_filenames = tar_filenames
        self.tar_filepaths = tar_filepaths
        self.part_filter = part_filter
        self.itar_files_cache = {}
        self.itar_cache_size = itar_cache_size
        self.sample_filter = sample_filter

    @abstractmethod
    def __len__(self) -> int:
        """Returns the total number of samples in the reader."""
        raise NotImplementedError

    @abstractmethod
    def __str__(self) -> str:
        """
        Must return a descriptive string of the concrete reader.
        """
        raise NotImplementedError

    @abstractmethod
    def _get_itar_sample_pointer(self, idx: T_index) -> ITarSamplePointer:
        """Get the ITarSample object for the given index."""
        raise NotImplementedError

    def _get_itarfile_cached(self, tar_file_id: int) -> ITarFile:
        """
        Get the ITarFile object for the given tar file id.
        If the file is not already open, open it. If we exceed
        the global cache limit, close the least recently used file.
        """
        if tar_file_id not in self.itar_files_cache:
            file_object = self.tar_filepaths[tar_file_id].open(mode="rb")
            tar_file = ITarFile.open(fileobj=file_object, mode="r:")
            self.itar_files_cache[tar_file_id] = tar_file

        # If we hit the limit of open files, close the least recently used file
        while len(self.itar_files_cache) > self.itar_cache_size:
            # Get the oldest file
            lru_key = next(iter(self.itar_files_cache))

            self.itar_files_cache[lru_key].fileobj.close()
            self.itar_files_cache[lru_key].close()
            del self.itar_files_cache[lru_key]

        return self.itar_files_cache[tar_file_id]

    def _get_item_by_sample_pointer(
        self,
        sample_pointer: ITarSamplePointer,
        restore_index: Union[str, int],
        entry_match_fn: Optional[Callable[[str], bool]] = None,
    ) -> Union["ITarReader", FilteredSample, None]:
        """
        Get a sample from the dataset or slice it.

        Args:
            sample_pointer: The sample pointer to get the sample from.
            sample_index: The global index of the sample in the dataset.
            entry_match_fn: An optional function to filter the entries in the sample.

        Returns:
            The sample or None if the sample is not found.
        """

        # Open the tar file (cached)
        tar_file = self._get_itarfile_cached(sample_pointer.tar_file_id)
        shard_name = self.tar_filenames[sample_pointer.tar_file_id]
        sample_base_name = None
        sample_name = None
        group_parts: Dict[str, bytes] = {}
        file_names: list[str] = []

        # Position the tar file at the correct offset
        tar_file.offset = sample_pointer.byte_offset

        while tar_file.offset < sample_pointer.byte_offset + sample_pointer.byte_size:
            tarinfo = tar_file.next()
            if tarinfo is None:
                raise ValueError(
                    f"Unexpected end of tar file: {self.tar_filenames[sample_pointer.tar_file_id]}"
                )
            fname = tarinfo.name
            if not tarinfo.isfile() or fname is None:
                continue
            if skip_meta_re.match(fname):
                continue

            # Extract the base_name and extension
            m = split_name_re.match(fname)
            if not m:
                continue
            cur_base_name, cur_ext = m.groups()

            if sample_base_name is None:
                sample_base_name = cur_base_name
                sample_name = f"{shard_name}/{cur_base_name}"
                if self.sample_filter is not None and not self.sample_filter(sample_name):
                    return None
            else:
                if sample_base_name != cur_base_name:
                    raise ValueError(
                        f"Inconsistent sample base name: {sample_base_name} vs {cur_base_name}"
                    )

            if entry_match_fn is not None:
                # If entry_match_fn is provided, use it to determine if we should take this entry
                take_entry = entry_match_fn(fname)
            else:
                # If no entry_match_fn is provided, use the part_filter to determine if we should take this entry
                take_entry = self.part_filter is None or self.part_filter(cur_ext)

            if take_entry:
                member_bytes = tar_file.extractfile(tarinfo).read()
                group_parts[cur_ext] = member_bytes
                file_names.append(fname)
        if sample_base_name is None:
            raise ValueError(f"No valid files found in sample {sample_pointer}")

        return FilteredSample(
            __key__=f"{shard_name}/{sample_base_name}",
            __shard__=self.tar_filenames[sample_pointer.tar_file_id],
            __restore_key__=("Webdataset", restore_index),
            __sources__=(
                SourceInfo(
                    dataset_path=self.base_path,
                    index=restore_index,
                    shard_name=shard_name,
                    file_names=tuple(file_names),
                ),
            ),
            **group_parts,
        )

    @overload
    def __getitem__(self, key: T_index) -> Optional[FilteredSample]: ...

    @overload
    def __getitem__(self, key: slice) -> "ITarReader": ...

    def __getitem__(self, key: Union[slice, T_index]) -> Union["ITarReader", FilteredSample, None]:
        """
        Get a sample from the dataset or slice it.
        """

        if isinstance(key, slice):
            # Return a new reader with a sliced samples tensor
            raise NotImplementedError("Slicing is not yet implemented")
        elif isinstance(key, int):
            idx = key
        else:
            raise TypeError("Invalid argument type for __getitem__")

        sample_pointer = self._get_itar_sample_pointer(idx)

        return self._get_item_by_sample_pointer(sample_pointer, idx)


class JoinIndexFileITarReader(ITarReader[int]):
    """
    A concrete ITarReader that reads samples from a join index file (via JoinIndexReader).
    """

    index_file: EPath
    column: int
    index_reader_cache: Dict[int, JoinIndexReader]
    index_reader_cache_size: int

    def __init__(
        self,
        index_file: EPath,
        column: int,
        tar_filenames: List[str],
        base_path: EPath,
        part_filter: Optional[Callable[[str], bool]] = None,
        itar_cache_size: int = 5,
        sample_filter: Optional[Callable[[str], bool]] = None,
    ):
        self.index_file = index_file
        self.column = column

        # Create the full path to each tar file
        tar_filepaths = [base_path / fn for fn in tar_filenames]

        self.index_reader_cache = {}
        self.index_reader_cache_size = itar_cache_size

        super().__init__(
            base_path=base_path,
            tar_filenames=tar_filenames,
            tar_filepaths=tar_filepaths,
            part_filter=part_filter,
            itar_cache_size=itar_cache_size,
            sample_filter=sample_filter,
        )

    def _get_join_index_reader_cached(self, sample_idx: int) -> JoinIndexReader:
        """
        Get the JoinIndexReader object for the given sample index, or create it if it doesn't exist.
        """

        if sample_idx not in self.index_reader_cache:
            index_reader = JoinIndexReader(self.index_file, column=self.column)
            self.index_reader_cache[sample_idx] = index_reader

        # If we hit the limit of open files, close the least recently used file
        while len(self.index_reader_cache) > self.index_reader_cache_size:
            # Get the oldest file
            lru_key = next(iter(self.index_reader_cache))

            self.index_reader_cache[lru_key].close()
            del self.index_reader_cache[lru_key]

        return self.index_reader_cache[sample_idx]

    def _get_itar_sample_pointer(self, sample_idx: int) -> ITarSamplePointer:
        """
        Get the ITarSample object for the given index.
        """
        index_reader = self._get_join_index_reader_cached(sample_idx)
        row = index_reader[sample_idx]

        # Update cache entry
        new_offset = index_reader.tell_row()
        del self.index_reader_cache[sample_idx]
        self.index_reader_cache[new_offset] = index_reader

        assert len(row) == 1
        shard_idx, byte_offset, byte_size = row[0]

        return ITarSamplePointer(
            tar_file_id=shard_idx,
            byte_offset=byte_offset,
            byte_size=byte_size,
        )

    def __len__(self) -> int:
        try:
            # Get any reader, they will all work
            index_reader = next(iter(self.index_reader_cache.values()))
        except StopIteration:
            # If there's no reader yet, we need to create one to get the length
            index_reader = self._get_join_index_reader_cached(0)

        return len(index_reader)

    def __str__(self) -> str:
        return (
            f"JoinIndexFileITarReader("
            f"len={len(self)}, base_path={self.base_path}, "
            f"len(shards)={len(self.tar_filenames)}, "
            f"shards=[{self.tar_filenames[0] if self.tar_filenames else 'N/A'}, ...])"
        )


class ShardInfosITarReader(ITarReader[int]):
    """
    A concrete ITarReader that constructs its internal sample list from a list of ShardInfos.
    """

    shard_infos: List[ShardInfo]
    shard_tar_file_idxs: List[int]
    shard_count_cumsum: List[int]
    cached_offset_reader: CachedItarOffsetReader

    def __init__(
        self,
        base_path: EPath,
        shard_infos: List[ShardInfo],
        part_filter: Optional[Callable[[str], bool]] = None,
        itar_cache_size: int = 5,
        sample_filter: Optional[Callable[[str], bool]] = None,
    ):
        # Build the tar_filenames and tar_filepaths from shard_infos,
        # constructing the samples tensor as we go.
        cur_tar_files: Dict[str, Tuple[int, EPath]] = {}

        self.shard_infos = shard_infos

        # Compute the cumsum of the shard counts, so that we can look up
        # the shard index for a given sample index.
        # Get all tar files from the shard_infos

        self.shard_count_cumsum = [0]
        self.shard_tar_file_idxs = []
        sample_idx = 0
        for shardinfo in shard_infos:
            filepath = shardinfo.path
            filename = shardinfo.name

            if filename not in cur_tar_files:
                cur_tar_files[filename] = (len(cur_tar_files), filepath)

            sample_idx += shardinfo.count
            self.shard_count_cumsum.append(sample_idx)
            self.shard_tar_file_idxs.append(cur_tar_files[filename][0])

        tar_filenames = list(cur_tar_files.keys())
        tar_filepaths = [p[1] for p in cur_tar_files.values()]

        # Instantiate cached reader for the .tar.idx files
        self.cached_offset_reader = CachedItarOffsetReader(cache_size=itar_cache_size)

        super().__init__(
            base_path=base_path,
            tar_filenames=tar_filenames,
            tar_filepaths=tar_filepaths,
            part_filter=part_filter,
            itar_cache_size=itar_cache_size,
            sample_filter=sample_filter,
        )

    def _get_itar_sample_pointer(self, idx: int) -> ITarSamplePointer:
        """
        Get the ITarSample object for the given index.
        """

        # Find the shard index using binary search
        shard_idx = bisect_right(self.shard_count_cumsum, idx) - 1
        if shard_idx < 0 or shard_idx >= len(self.shard_infos):
            raise IndexError(f"Index out of bounds: {idx}")

        # Get the shard info for the given index
        shard = self.shard_infos[shard_idx]
        sample_idx_in_shard_file = idx - self.shard_count_cumsum[shard_idx]

        # Now we know the tar file and the sample offset in the file.
        # We need to figure out the byte offset and size of the sample,
        # by looking it up in the .tar.idx file.
        byte_offset, byte_size = self.cached_offset_reader.get_itar_byte_offset(
            shard.path, sample_idx_in_shard_file
        )

        return ITarSamplePointer(
            tar_file_id=self.shard_tar_file_idxs[shard_idx],
            byte_offset=byte_offset,
            byte_size=byte_size,
        )

    def __len__(self) -> int:
        return self.shard_count_cumsum[-1]

    def __str__(self) -> str:
        return (
            f"ShardInfosITarReader("
            f"len={len(self)}, base_path={self.base_path}, "
            f"len(shards)={len(self.tar_filenames)}, "
            f"shards=[{self.tar_filenames[0] if self.tar_filenames else 'N/A'}, ...])"
        )


class SqliteITarEntryReader(ITarReader[str]):
    """
    A concrete ITarReader that constructs its internal sample list from a SQLite database.
    """

    sqlite_reader: SqliteIndexReader

    def __init__(
        self,
        base_path: EPath,
        part_filter: Optional[Callable[[str], bool]] = None,
        itar_cache_size: int = 5,
        sample_filter: Optional[Callable[[str], bool]] = None,
        key_is_full_entryname: bool = False,
    ):
        from megatron.energon.flavors.webdataset.config import MAIN_FOLDER_NAME
        from megatron.energon.flavors.webdataset.indexing import SqliteIndexReader

        # shard_name_to_info_idx = {name: i for i, name in enumerate(wds_meta.info_shard_files)}
        tar_filenames = get_info_shard_files(base_path)
        tar_filepaths = [base_path / fn for fn in tar_filenames]

        # Initialize the SQLite reader
        sqlite_path = base_path / MAIN_FOLDER_NAME / "index.sqlite"
        self.sqlite_reader = SqliteIndexReader(sqlite_path)

        self.key_is_full_entryname = key_is_full_entryname

        super().__init__(
            base_path=base_path,
            tar_filenames=tar_filenames,
            tar_filepaths=tar_filepaths,
            part_filter=part_filter,
            itar_cache_size=itar_cache_size,
            sample_filter=sample_filter,
        )

    def _get_itar_sample_pointer(self, sample_key: str) -> ITarSamplePointer:
        """
        Get the ITarSample object for the given index.
        """

        return self.sqlite_reader.get_sample_pointer_by_key(sample_key)

    def list_all_samples(self) -> Generator[Tuple[str, int, int], None, None]:
        return self.sqlite_reader.list_all_samples()

    def list_all_sample_parts(self) -> Generator[Tuple[str, int, int], None, None]:
        return self.sqlite_reader.list_all_sample_parts()

    def list_sample_parts(
        self, sample_key: str, slow_mode: bool = False
    ) -> Generator[Tuple[str, int, int], None, None]:
        """Given a sample key, list all its parts. (E.g. given 001, list 001.jpg, 001.json, etc.)

        If allow_fallback is True, and the database is an older version, which
        does not contain the sample_parts table, we will try to find the sample_parts
        in the tar files.

        Args:
            sample_key: The sample key to list the parts of.
            allow_fallback: If True, and the database is an older version, which
                does not contain the sample_parts table, we will try to find the sample_parts
                in the tar files.

        Returns:
            A generator of tuples of (part_name, size, tar_file_id)
        """

        if not slow_mode:
            yield from self.sqlite_reader.list_sample_parts(sample_key)
        else:
            sample_pointer = self._get_itar_sample_pointer(sample_key)

            sample = self._get_item_by_sample_pointer(sample_pointer, 0, entry_match_fn=None)
            assert isinstance(sample, dict), f"Sample not found: {sample_pointer}"

            for ext in sample.keys():
                if not ext.startswith("__"):
                    yield ext, len(sample[ext]), sample_pointer.tar_file_id

    def get_total_size(self) -> int:
        return self.sqlite_reader.get_total_size()

    @overload
    def __getitem__(self, key: str) -> Union[FilteredSample, tuple[bytes, SourceInfo]]: ...

    @overload
    def __getitem__(self, key: slice) -> "ITarReader": ...

    def __getitem__(
        self, key: Union[slice, str]
    ) -> Union[FilteredSample, tuple[bytes, SourceInfo], ITarReader]:
        """
        Either get a sample from the dataset by the sample key including all its entries,
        or get the bytes of a specific entry by the full filename of the entry inside the tar.
        """

        if isinstance(key, slice):
            # Return a new reader with a sliced samples tensor
            raise NotImplementedError("Slicing is not yet implemented")
        assert isinstance(key, str), "Invalid argument type for __getitem__"

        if self.key_is_full_entryname:
            m = split_name_re.match(key)
            if not m:
                raise ValueError(f"Invalid file name: {key}")

            sample_key, sample_ext = m.groups()
            entry_match_fn = lambda fname: key == fname
        else:
            sample_key = key
            sample_ext = None
            entry_match_fn = None

        sample_pointer = self._get_itar_sample_pointer(sample_key)

        sample = self._get_item_by_sample_pointer(
            sample_pointer, key, entry_match_fn=entry_match_fn
        )
        assert sample is not None, f"Sample not found: {sample_key}"

        if self.key_is_full_entryname:
            assert isinstance(sample_ext, str)
            assert len(sample["__sources__"]) == 1
            # Return the bytes directly
            return sample[sample_ext], sample["__sources__"][0]
        else:
            return sample  # Return the FilteredSample

    def __len__(self) -> int:
        """Return the total number of samples in the database."""
        return self.sqlite_reader.get_sample_count()

    def __str__(self) -> str:
        """Return a descriptive string of this reader."""
        return (
            f"SqliteITarEntryReader("
            f"len={len(self)}, base_path={self.base_path}, "
            f"len(shards)={len(self.tar_filenames)}, "
            f"shards=[{self.tar_filenames[0] if self.tar_filenames else 'N/A'}, ...])"
        )

    def close(self):
        """Close the SQLite reader and any open ITarFiles."""
        # Close the SQLite reader
        if hasattr(self, "sqlite_reader") and self.sqlite_reader is not None:
            self.sqlite_reader.close()

        # Close any open ITarFiles (using parent class implementation)
        for tar_file_id in list(self.itar_files_cache.keys()):
            tar_file = self.itar_files_cache[tar_file_id]
            if (
                tar_file is not None
                and hasattr(tar_file, "fileobj")
                and tar_file.fileobj is not None
            ):
                tar_file.fileobj.close()
            if tar_file is not None and hasattr(tar_file, "close"):
                tar_file.close()
            del self.itar_files_cache[tar_file_id]



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from enum import Enum
from typing import TypeVar

import PIL.Image
import torch

TTransform = TypeVar("TTransform", bound=torch.nn.Module)


class Sampling(Enum):
    BILINEAR = PIL.Image.BILINEAR
    NEAREST = PIL.Image.NEAREST



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import contextlib
import inspect
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterator,
    List,
    Optional,
    TypeVar,
    Union,
)

from megatron.energon.errors import SYSTEM_EXCEPTIONS, FatalSampleError
from megatron.energon.flavors.base_dataset import (
    SavableDataset,
    add_sample_restore_key,
    set_sample_restore_key,
)
from megatron.energon.source_info import SourceInfo
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers._log_exception import log_exception
from megatron.energon.wrappers.base import BaseWrapperDataset, SampleIndex, get_sample_restore_key
from megatron.energon.wrappers.buffer import SavableSampleBuffer
from megatron.energon.wrappers.skip import SkipSample

T_sample = TypeVar("T_sample")
T_encoded_sample = TypeVar("T_encoded_sample")
T_batch_sample = TypeVar("T_batch_sample")


class PackingDataset(
    BaseWrapperDataset[T_sample, T_encoded_sample, T_batch_sample],
    Generic[T_sample, T_encoded_sample, T_batch_sample],
):
    """This dataset wrapper transforms samples of a dataset into chunks/packs of samples, which are
    then combined into a batch."""

    buffer_size: int
    pre_packer: Callable[[List[T_sample]], List[List[T_sample]]]
    sample_encoder: Optional[Callable[[T_sample], T_encoded_sample]]
    sample_encoder_stateless: bool
    final_packer: Callable[[List[T_encoded_sample]], T_batch_sample]
    final_packer_stateless: bool
    packer_config: Optional[Union[Dict[str, Any], Callable[[], Dict[str, Any]]]]
    error_handler: Callable[[Exception, List[T_sample], list[SourceInfo]], None]

    #: The buffer for collecting the samples that shall be packed.
    _reading_buffer: SavableSampleBuffer

    #: Contains the pre-selected samples to be packed.
    #: The full buffer will be passed to the pre_packer.
    _pre_packing_buffer: SavableSampleBuffer

    #: Lengths of the selected groups of samples to be packed together.
    #: The samples are stored sequentially in the pre_packing_buffer because
    #: SavableSampleBuffer doesn't support nesting. But to keep the groups
    #: separate, we need to store the lengths of the groups here.
    _pre_packing_lengths: List[List[int]]

    #: Sample index for the pre_packer
    _pre_packing_sample_index: SampleIndex

    #: Sample index for the sample_encoder
    _sample_encoder_sample_index: SampleIndex

    #: Sample index for the final_packer
    _final_packing_sample_index: SampleIndex

    _savable_fields = (
        "_reading_buffer",
        "_pre_packing_buffer",
        "_pre_packing_lengths",
        "_pre_packing_sample_index",
        "_sample_encoder_sample_index",
        "_final_packing_sample_index",
    )

    def __init__(
        self,
        dataset: SavableDataset[T_sample],
        buffer_size: int,
        pre_packer: Callable[[List[T_sample]], List[List[T_sample]]],
        final_packer: Callable[[List[T_encoded_sample]], T_batch_sample],
        *,
        final_packer_stateless: bool = False,
        sample_encoder: Optional[Callable[[List[T_sample]], T_encoded_sample]] = None,
        sample_encoder_stateless: bool = False,
        packer_config: Optional[Union[Dict[str, Any], Callable[[], Dict[str, Any]]]] = None,
        error_handler: Callable[
            [Exception, List[T_sample], list[SourceInfo]], None
        ] = log_exception,
        pre_packer_failure_tolerance: Optional[int] = 100,
        final_packer_failure_tolerance: Optional[int] = 100,
        sample_encoder_failure_tolerance: Optional[int] = 100,
        worker_config: WorkerConfig,
    ):
        """Construct a PackingDataset which is used for sequence packing.
        Using a pre_packer and final_packer, it buffers the incoming samples, groups
        them together based on the logic provided by the pre_packer, and then (using
        the final_packer) combines each group into a packed single sample also called
        a "pack" or a "packed sequence".

        Args:
            dataset: The input dataset to wrap
            buffer_size: The desired size of the input buffer for pre packing. Last buffer of a dataset may be smaller.
            pre_packer: Function which selects samples from the buffer to be packed together.
                May raise :exc:`megatron.energon.SkipSample` to skip a buffer.
            final_packer: Function which combines the selected samples into a single sample.
            final_packer_stateless: If True, the final_packer is stateless, thus samples can be
                stored/restored.
            sample_encoder: Function which encodes the samples.
            sample_encoder_stateless: If True, the sample_encoder is stateless, thus samples can be
                stored/restored.
            packer_config: Configuration for the (pre|final)_packer functions. If callable, it should return the
                configuration. Defaults to None.
            error_handler: Function which handles exceptions raised by the batcher. The default
                implementation logs the exception.
            pre_packer_failure_tolerance: Maximum number of pre-packer failures before raising an error.
            final_packer_failure_tolerance: Maximum number of final-packer failures before raising an error.
            sample_encoder_failure_tolerance: Maximum number of sample-encoder failures before raising an error.
            worker_config: Configuration for the workers.
        """
        super().__init__(dataset, worker_config=worker_config)

        assert buffer_size > 0, "Packing buffer size must be greater than 0."

        self.buffer_size = buffer_size
        self.pre_packer = pre_packer
        self.final_packer = final_packer
        self.final_packer_stateless = final_packer_stateless
        self.sample_encoder = sample_encoder
        self.sample_encoder_stateless = True if sample_encoder is None else sample_encoder_stateless
        self.packer_config = packer_config
        self.error_handler = error_handler

        self.pre_packer_failure_tolerance = pre_packer_failure_tolerance
        self.final_packer_failure_tolerance = final_packer_failure_tolerance
        self.sample_encoder_failure_tolerance = sample_encoder_failure_tolerance

        self.reset_state_own()

    def reset_state_own(self) -> None:
        self._reading_buffer = SavableSampleBuffer(self.dataset, worker_config=self.worker_config)
        self._pre_packing_buffer = SavableSampleBuffer(
            self.dataset, worker_config=self.worker_config
        )
        self._pre_packing_lengths = []
        self._pre_packing_sample_index = SampleIndex(self.worker_config, src=self)
        self._final_packing_sample_index = SampleIndex(self.worker_config, src=self)
        self._sample_encoder_sample_index = SampleIndex(self.worker_config, src=self)

    def __len__(self):
        """The real length is unknown, since it depends on the packing function.
        We approximate it by the length of the source dataset."""

        return len(self.dataset)

    def _fill_reading_buffer(self, source_iter: Iterator, log_progress: bool = False) -> bool:
        """
        Fill the reading buffer with samples from the dataset source iterator.

        Args:
            source_iter: Iterator of samples from the dataset.
            log_progress: If True, log the progress of the filling.

        Returns:
            True if samples are successfully read into the buffer, False if no more data.
        """

        if log_progress:
            import tqdm

            pbar_ctx = pbar = tqdm.tqdm(total=self.buffer_size, desc="Filling reading buffer")
        else:
            pbar_ctx = contextlib.nullcontext()
            pbar = None

        with pbar_ctx:
            while len(self._reading_buffer) + len(self._pre_packing_buffer) < self.buffer_size:
                try:
                    sample = next(source_iter)
                    self._reading_buffer.append(sample)
                    if pbar is not None:
                        pbar.update(1)
                except StopIteration:
                    return False
        return True

    def __iter__(self) -> Iterator[T_batch_sample]:
        pre_packing_lengths = self._pre_packing_lengths
        # The source dataset
        src_iter = iter(self.dataset)

        last_pre_pack_failures = 0
        last_final_pack_failures = 0
        last_sample_encoder_failures = 0

        self._pre_packing_buffer.worker_start()
        self._reading_buffer.worker_start()

        is_initial_pack = True

        def encode_pack_samples(pack: List[T_sample]) -> List[T_encoded_sample]:
            """Encode the samples in the pack using the sample encoder."""
            nonlocal last_sample_encoder_failures

            # Apply the sample encoder to the pack
            if self.sample_encoder is None:
                return pack
            encoded_pack = []
            for sample in pack:
                try:
                    with self._sample_encoder_sample_index.ctx() as encode_idx:
                        encoded_sample = self.sample_encoder(sample)
                    assert not isinstance(encoded_sample, Generator), "Generator not supported"
                    encoded_pack.append(
                        add_sample_restore_key(
                            encoded_sample,
                            encode_idx,
                            src=self,
                        )
                    )
                except SkipSample:
                    pass
                except SYSTEM_EXCEPTIONS:
                    raise FatalSampleError.from_sample(pack)
                except Exception as e:
                    self.error_handler(e, [sample])
                    last_sample_encoder_failures += 1
                    if (
                        self.sample_encoder_failure_tolerance is not None
                        and last_sample_encoder_failures >= self.sample_encoder_failure_tolerance
                    ):
                        raise FatalSampleError.from_sample(
                            pack,
                            f"Sample encoder {self.sample_encoder} failed {last_sample_encoder_failures} times. Likely your code or dataset are broken.",
                        )
            return encoded_pack

        def next_pre_pack():
            """Take the samples from the reading buffer and select groups of samples to be packed
            together."""
            nonlocal last_pre_pack_failures

            assert len(self._pre_packing_buffer) == 0
            if len(self._reading_buffer) > 0:
                # Take all samples from the reading buffer and pre_pack them
                samples = list(self._reading_buffer)
                # Clear buffer and pre_packing_lengths
                self._reading_buffer.clear()
                pre_packing_lengths.clear()
                # Now pre pack the samples
                try:
                    with self._pre_packing_sample_index.ctx():
                        pre_packs = self.pre_packer(samples)
                except SkipSample:
                    pre_packs = []
                except SYSTEM_EXCEPTIONS:
                    raise FatalSampleError.from_sample(samples)
                except Exception as e:
                    self.error_handler(e, samples)
                    pre_packs = []
                    last_pre_pack_failures += 1
                    if (
                        self.pre_packer_failure_tolerance is not None
                        and last_pre_pack_failures >= self.pre_packer_failure_tolerance
                    ):
                        raise FatalSampleError.from_sample(
                            samples,
                            f"Pre packer {self.pre_packer} failed {last_pre_pack_failures} times. Likely your code or dataset are broken.",
                        )

                # Put the pre-packed samples into the pre_packing_buffer
                # They will be flattened here to avoid nested buffers
                # But the lengths of the groups are stored in pre_packing_lengths
                # so that the groups can be separated later
                for pre_pack in pre_packs:
                    if len(pre_pack) > 0:
                        self._pre_packing_buffer.extend(pre_pack)
                        pre_packing_lengths.append(len(pre_pack))

        def next_final_pack() -> Generator[T_batch_sample, None, None]:
            """Yield the next packs from the buffer. The final packer is called on the fly."""
            nonlocal last_final_pack_failures

            pack = list(self._pre_packing_buffer[: pre_packing_lengths[0]])
            if len(pack) == 0:
                return
            pack = encode_pack_samples(pack)

            del self._pre_packing_buffer[: pre_packing_lengths[0]]
            del pre_packing_lengths[0]
            try:
                pack_restore_keys = tuple(get_sample_restore_key(sample) for sample in pack)
                with self._final_packing_sample_index.ctx() as pack_idx:
                    final_packed_sample = self.final_packer(pack)
                if isinstance(final_packed_sample, Generator):
                    assert inspect.isgeneratorfunction(self.final_packer), (
                        f"Generator in {self.final_packer} but not marked as such."
                    )
                    for pack_sub_idx, (pack_idx, inner_batch_sample) in enumerate(
                        self._final_packing_sample_index.iter_ctx(final_packed_sample, pack_idx)
                    ):
                        yield set_sample_restore_key(
                            inner_batch_sample,
                            pack_idx,
                            pack_sub_idx,
                            *pack_restore_keys,
                            src=self,
                        )
                else:
                    yield set_sample_restore_key(
                        final_packed_sample,
                        pack_idx,
                        *pack_restore_keys,
                        src=self,
                    )
            except SkipSample:
                pass
            except SYSTEM_EXCEPTIONS:
                raise FatalSampleError.from_sample(pack)
            except Exception as e:
                self.error_handler(e, pack)
                last_final_pack_failures += 1
                if (
                    self.final_packer_failure_tolerance is not None
                    and last_final_pack_failures >= self.final_packer_failure_tolerance
                ):
                    raise FatalSampleError.from_sample(
                        pack,
                        f"Final packer {self.final_packer} failed {last_final_pack_failures} times. Likely your code or dataset are broken.",
                    )

        # Main loop:
        pre_pack_round = 0
        while True:
            if pre_pack_round > self.pre_packer_failure_tolerance:
                raise RuntimeError(
                    f"Pre packer {self.pre_packer} did not yield any packs after {pre_pack_round} rounds. Likely your code or dataset are broken."
                )
            # Fill a portion of the buffer
            if not self._fill_reading_buffer(src_iter, log_progress=is_initial_pack):
                # Break out of the main loop when the source is exhausted.
                break
            is_initial_pack = False

            # Create new pre packs if necessary
            if len(pre_packing_lengths) == 0:
                assert len(self._pre_packing_buffer) == 0
                assert len(self._reading_buffer) == self.buffer_size
                next_pre_pack()
                if len(pre_packing_lengths) == 0:
                    # Retry packing, nothing was returned.
                    pre_pack_round += 1
                    continue

            if len(pre_packing_lengths) > 0:
                pre_pack_round = 0

            yield from next_final_pack()

        # Yield the remaining packs, flushing the collecting buffer
        while len(pre_packing_lengths) > 0:
            yield from next_final_pack()

        # If there are still samples in the partial reading buffer, pre-pack them and yield the
        # resulting (partial) packs
        if len(self._reading_buffer) > 0:
            next_pre_pack()

        # Yield the remaining packs, flushing the collecting buffer
        while len(pre_packing_lengths) > 0:
            yield from next_final_pack()

    def can_restore_sample(self) -> bool:
        # Cannot really verify if the returned elements contain a __restore_key__.
        # If the user wants to use this, well...
        return (
            super().can_restore_sample()
            and self.final_packer_stateless
            and self.sample_encoder_stateless
        )

    def assert_can_restore(self):
        assert self.final_packer_stateless and self.sample_encoder_stateless, (
            f"Final packer {self.final_packer} and sample encoder {self.sample_encoder} must be stateless to restore samples."
        )
        super().assert_can_restore()

    def restore_sample(self, restore_key: Any) -> T_sample:
        # We need to store multiple indices to restore a batch.
        self.assert_can_restore()
        if inspect.isgeneratorfunction(self.final_packer):
            id, pack_idx, pack_sub_idx, *pack_restore_keys = restore_key
            id, pack_idx, pack_sub_idx, *pack_restore_keys = restore_key
            assert id == type(self).__name__
        else:
            id, pack_idx, *pack_restore_keys = restore_key
            id, pack_idx, *pack_restore_keys = restore_key
            assert id == type(self).__name__

        pack = []
        for inner_idx in pack_restore_keys:
            if self.sample_encoder is not None:
                id, sample_idx, *inner_idx = inner_idx
                assert id == type(self).__name__
                id, sample_idx, *inner_idx = inner_idx
                assert id == type(self).__name__
                assert isinstance(sample_idx, int)
            sample = self.dataset.restore_sample(inner_idx)
            if self.sample_encoder is not None:
                with self._sample_encoder_sample_index.ctx(sample_idx):
                    sample = self.sample_encoder(sample)
                assert not isinstance(sample, Generator), "Generator not supported"
                sample = add_sample_restore_key(sample, sample_idx, src=self)
            pack.append(sample)
        with self._final_packing_sample_index.ctx(pack_idx):
            final_pack = self.final_packer(pack)
        if isinstance(final_pack, Generator):
            assert inspect.isgeneratorfunction(self.final_packer), (
                f"Generator in {self.final_packer} but not marked as such."
            )
            for cur_batch_sub_idx, (pack_idx, inner_batch_sample) in enumerate(
                self._final_packing_sample_index.iter_ctx(final_pack, pack_idx)
            ):
                if cur_batch_sub_idx == pack_sub_idx:
                    return set_sample_restore_key(
                        inner_batch_sample,
                        pack_idx,
                        pack_sub_idx,
                        *pack_restore_keys,
                        src=self,
                    )
            assert False, f"Pack sub-index {pack_sub_idx} not found in pack"
        else:
            return set_sample_restore_key(final_pack, pack_idx, *pack_restore_keys, src=self)

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "buffer_size": self.buffer_size,
            "pre_packer": self._function_config(self.pre_packer),
            "final_packer": self._function_config(self.final_packer),
            "final_packer_stateless": self.final_packer_stateless,
            **(
                {
                    "packer_config": (
                        self.packer_config() if callable(self.packer_config) else self.packer_config
                    )
                }
                if self.packer_config
                else {}
            ),
            "error_handler": self._function_config(self.error_handler),
            "worker_config": self.worker_config.config(),
            "dataset": self.dataset.config(),
        }

    def __str__(self):
        return f"PackingDataset(buffer_size={self.buffer_size}, pre_packer={self.pre_packer}, final_packer={self.final_packer}, dataset={self.dataset})"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional

import torch

from megatron.energon.edataclass import edataclass
from megatron.energon.epathlib.epath import EPath
from megatron.energon.errors import warn_deprecated
from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.webdataset import DefaultDecoderWebdatasetFactory


@edataclass
class ImageClassificationSample(Sample):
    """Sample type for classifying an image."""

    #: The input image tensor in the shape (C, H, W)
    image: torch.Tensor
    #: The class label of the image
    label: Optional[int] = None
    #: The class label of the image
    label_name: Optional[str] = None


class ImageClassificationWebdataset(DefaultDecoderWebdatasetFactory[ImageClassificationSample]):
    __sample_type__ = ImageClassificationSample

    def __init__(self, path: EPath, **kwargs):
        warn_deprecated(
            f"{type(self)} is deprecated, use the default instead and set the sample_type:\n"
            f"To convert, update your {path}/.nv-meta/dataset.yaml to:\n"
            f"# remove top-level __module__ and __class__\n"
            f"sample_type:\n"
            f"  __module__: megatron.energon\n"
            f"  __class__: {self.__sample_type__.__name__}\n"
            f"# Keep the remaining content"
        )
        super().__init__(path, **kwargs)



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Callable, Dict, Generic, Iterator, Optional, TypeVar, Union

from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers.base import BaseWrapperDataset, SampleIndex

T_sample = TypeVar("T_sample")


class FilterDataset(BaseWrapperDataset[T_sample, T_sample], Generic[T_sample]):
    """This dataset wrapper applies a custom filter function to each sample and does not yield
    filtered samples."""

    filter_fn: Callable[[T_sample], bool]
    filter_fn_config: Optional[Union[Dict[str, Any], Callable[[], Dict[str, Any]]]]
    _sample_index: SampleIndex

    _savable_fields = ("_sample_index",)

    def __init__(
        self,
        dataset: SavableDataset[T_sample],
        *,
        filter_fn: Callable[[T_sample], bool],
        filter_fn_config: Optional[Union[Dict[str, Any], Callable[[], Dict[str, Any]]]] = None,
        worker_config: WorkerConfig,
    ):
        """Construct a MapDataset.

        Args:
            dataset: The input dataset to wrap
            filter_fn: The function to apply to each sample. If it returns `True`, the sample is
               accepted.
            filter_fn_config: Configuration for the filter function. If callable, it should return the
                configuration. Defaults to None.
            worker_config: Configuration for the workers.
        """
        super().__init__(dataset, worker_config=worker_config)
        self.filter_fn = filter_fn
        self.filter_fn_config = filter_fn_config

        self.reset_state_own()

    def reset_state_own(self) -> None:
        self._sample_index = SampleIndex(self.worker_config, src=self)

    def __len__(self):
        return len(self.dataset)

    def __iter__(self) -> Iterator[T_sample]:
        for sample in self.dataset:
            with self._sample_index.ctx():
                filter_res = self.filter_fn(sample)
            if filter_res:
                yield sample

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset": self.dataset.config(),
            "filter_fn": self._function_config(self.filter_fn),
            **(
                {
                    "filter_fn_config": (
                        self.filter_fn_config()
                        if callable(self.filter_fn_config)
                        else self.filter_fn_config
                    )
                }
                if self.filter_fn_config
                else {}
            ),
        }

    def __str__(self):
        return f"FilterDataset(filter_fn={self.filter_fn}, dataset={self.dataset})"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Optional, Union

from megatron.energon.cache.base import FileStore, FileStoreDecoder
from megatron.energon.epathlib import EPath
from megatron.energon.flavors.webdataset.itar_reader import SqliteITarEntryReader
from megatron.energon.source_info import SourceInfo


class DecodeFileStore(FileStore[Any]):
    """Used to wrap a FileStore and decode the data on access."""

    def __init__(
        self,
        inner_reader: FileStore[bytes],
        *,
        decoder: FileStoreDecoder,
    ):
        """
        Args:
            inner_reader: The FileStore to wrap.
            decoder: The decoder to apply to every item read from the FileStore.
        """

        self.inner_reader = inner_reader
        self.decoder = decoder

    def __getitem__(self, fname: str) -> tuple[Any, SourceInfo]:
        data, source_info = self.inner_reader[fname]
        return self.decoder.decode(fname, data), source_info

    def get_path(self) -> str:
        return self.inner_reader.get_path()

    def __str__(self):
        return f"DecodeFileStore(inner_reader={self.inner_reader}, decoder={self.decoder})"


class SystemFileStore(FileStore[bytes]):
    """A FileStore that reads files directly from the file system."""

    def __init__(self, base_dir: Optional[Union[EPath, str]] = None):
        """
        Args:
            base_dir: The base directory to use for relative paths. If None, you should only pass
                absolute paths to __getitem__.
        """

        self.base_dir = EPath(base_dir) if base_dir is not None else None

    def __getitem__(self, key: str) -> tuple[bytes, SourceInfo]:
        # Construct the full path from the dataset path and the file key
        if self.base_dir is None:
            file_path = EPath(key)
        else:
            file_path = self.base_dir / key

        # Read and return the file contents as bytes
        with file_path.open("rb") as f:
            data = f.read()

        return data, SourceInfo(
            dataset_path=self.base_dir,
            index=None,
            shard_name=None,
            file_names=(key,),
        )

    def get_path(self) -> str:
        """Returns the path to the dataset."""
        return str(self.base_dir)

    def __str__(self):
        return f"SystemFileStore(base_dir={self.base_dir})"


class WebdatasetFileStore(SqliteITarEntryReader, FileStore[bytes]):
    """This dataset will directly read files from the dataset tar files from a prepared energon dataset."""

    def __init__(
        self,
        dataset_path: EPath,
    ):
        super().__init__(base_path=dataset_path, key_is_full_entryname=True)

    def get_path(self) -> str:
        return str(self.base_path)
